﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;
using System.Text.RegularExpressions;

// The form retriever icon was made by Freepik from www.flaticon.com.

namespace CodeRetrieverDataflow
{
    public partial class Form1 : Form
    {
        public ListBox Boxnum { get; private set; }
        public bool BoxFlag { get; private set; }

        // This is the constructor for the Form1 class [note for Harry!].
        public Form1()
        {
            InitializeComponent();

            // Setting focus to the dataflow retrieval area.
            retrieveText.Select();
        }


        // This is the code that sets up the flag that determines if the default
        // box type is old or new. The checkbox "New Box" is checked as the default
        // when the program is first run. If the user has a form that is not in the
        // new box format then they need to uncheck the checkbox. The code that
        // controls this starts in the routine "globalRenumButton_Click" 
        // around line 1880.
        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox1.Checked)
            {
                BoxFlag = true;
                
            }
            else
            {
                BoxFlag = false;
                MessageBox.Show("Switching to box1_1");
            }

            
                
            
        }


        /*
        # This reads the .RBW file and builds a list of FormBox objects, setting each object to the
        # box number and the width is not set by this routine [XML file is needed for this].
        #
        # The ListBox form_box_list_box is then populated with the box numbers.
        */
        
        private void rubySourceCodeFiledocumentBoxNumbersToolStripMenuItem_Click(object sender, EventArgs e)
        {
            string fileName;

            fileName = GetFileName(curDir, "RBW");

            if (!String.IsNullOrEmpty(fileName))
            {
                try
                {
                    curDir = Path.GetDirectoryName(fileName);
                    form_box_list_label.Text = string.Format("{0} Box List", Path.GetFileNameWithoutExtension(fileName).ToUpper());
                }
                catch (ArgumentException ex)
                {
                    MessageBox.Show(ex.ToString(), "Invalid Path parameter.", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                catch (PathTooLongException ex)
                {
                    MessageBox.Show(ex.ToString(), "The path parameter is longer than the system-defined maximum length.", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.ToString());
                }

                // Read the Ruby code file and build a list of boxes found in the 'Field Data' section. Since we are dealing with
                // a .RBW file, we will set the "Width" member to "NO XML". form_box_list holds objects of FormBox see CustomClasses
                // for structure.
                //formBoxList = BuildBoxArrayRbw(fileName);

                //formBoxListBox.Items.Clear();
                foreach (FormBox formBox in formBoxList) { formBoxListBox.Items.Add(formBox.Box); }
            }
        }
        
        /*
        # This reads the .XML file and builds a list of FormBox objects, setting each object
        # to the box number and the width in chars that each box can hold. If the width cannot be
        # determined because the font type is not correct, we insert "MAN" for the width.
        #
        # The main purpose of reading the XML file is for dataflow truncation.
        #
        # This XML file event handler fires off from user mouse clicks on the Read-XML menu tab. The user
        # selects the appropriate .ZIP file from the Open File Dialog window, the preferred/required method of
        # selecting the .XML file. The selected .zip file will be unzipped and we will obtain box info from the
        # doc.xml file. Once we get the box info this routine will delete the extracted doc.xml file which briefly
        # resides in the temp folder on the C: drive as returned from Path.GetTempPath().
        #
        # I left in the code that will also process a .xml file selected directly by the user with no unzipping required.
        #
        # The ListBox form_box_list_box is then populated with the box numbers.
        */
        private void xMLFiledocumentBoxNumbersAndBoxWidthsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            string fileName;
            bool unzip = false;

            fileName = GetFileName(curDir, "XML");

            if (!String.IsNullOrEmpty(fileName))
            {
                try
                {
                    curDir = Path.GetDirectoryName(fileName);
                    form_box_list_label.Text = string.Format("{0} Box List", Path.GetFileNameWithoutExtension(fileName).ToUpper());
                    formBoxListBox.Items.Clear();
                }
                catch (ArgumentException ex)
                {
                    MessageBox.Show(ex.ToString(), "Invalid Path parameter.", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                catch (PathTooLongException ex)
                {
                    MessageBox.Show(ex.ToString(), "The path parameter is longer than the system-defined maximum length.", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.ToString());
                }

                try
                {
                    // Look for a .zip extension in the name, if so, we need to first extract the XML file "DOC.XML":
                    if (Path.GetExtension(fileName).ToLower() == ".zip")
                    {
                        unzip = true;
                        fileName = ExtractXmlFileFromZipFile(fileName);

                        if (String.IsNullOrEmpty(fileName))
                        {
                            // ExtractXmlFileFromZipFile() will return a null value when the .zip file is for a document with no boxes.
                            return;
                        }
                    }
                }
                catch (ArgumentException ex)
                {
                    MessageBox.Show(ex.ToString(), "Invalid Path string.", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.ToString());
                }

                // Read the .XML file and build a list of boxes found in the 'Field Data' section of a Ruby code file. Since we are dealing with
                // an .XML file, we will set the "Width" member to either the number of characters each box can hold, when the
                // font type is 'correct', or "MAN" when the font is of a type we can not use to determine the number of characters.
                // formBoxList holds objects of FormBox see CustomClasses for structure.
                formBoxList = BuildBoxArrayXml(fileName);

                if (formBoxList == null)
                {
                    ProcessErrors("Error! Tell Harry!!", 14, string.Format("BuildBoxArrayXml(\"{0}\")  \n\n\n  Which was called from:\n  xMLFiledocumentBoxNumbersAndBoxWidthsToolStripMenuItem_Click()", fileName), null);
                    return;
                }

                // Now we delete the extracted XML file if we had a .zip file.
                if (unzip)
                {
                    // Going overboard here???
                    try
                    {
                        File.Delete(fileName);
                    }
                    catch (ArgumentException ex)
                    {
                        MessageBox.Show(ex.ToString(), string.Format("Can not delete {0}.", fileName), MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                    catch (DirectoryNotFoundException ex)
                    {
                        MessageBox.Show(ex.ToString(), string.Format("Can not delete {0}.", fileName), MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                    catch (IOException ex)
                    {
                        MessageBox.Show(ex.ToString(), string.Format("Can not delete {0}.", fileName), MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                    catch (NotSupportedException ex)
                    {
                        MessageBox.Show(ex.ToString(), string.Format("Can not delete {0}.", fileName), MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                    catch (UnauthorizedAccessException ex)
                    {
                        MessageBox.Show(ex.ToString(), string.Format("Can not delete {0}.", fileName), MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show(ex.ToString());
                    }
                }

                foreach (FormBox formBox in formBoxList) { formBoxListBox.Items.Add(formBox.Box); }
            }
        }

        private void individualToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (profile != "Individual")
            {
                // Entering Individual Profile.
                ChangeProfile("Individual");
            }
        }

        private void entityToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (profile != "Entity")
            {
                // Entering Entity Profile.
                ChangeProfile("Entity");
            }
        }

        private void fiduciaryToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (profile != "Fiduciary")
            {
                // Entering Fiduciary Profile.
                ChangeProfile("Fiduciary");
            }
        }

        private void preparerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (profile != "Preparer")
            {
                // Entering Preparer Profile.
                ChangeProfile("Preparer");
            }
        }

        private void equifaxToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (profile != "Equifax")
            {
                // Entering Equifax Profile.
                ChangeProfile("Equifax");
            }
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        #region Tag Lists - Selected Index Changed

        private void tag_list_SelectedIndexChanged(object sender, EventArgs e)
        {
            // Return if nothing is selected in the list.
            if (tagList.SelectedIndex != -1)
            {
                Match match;
                profileTag = ((String)tagList.SelectedItem).ToUpper();
                string mnemonic_copy = "";

                if (!tagListLocked)
                {
                    // Clear the other lists.
                    subTagList.Items.Clear();
                    subTagCombList.Items.Clear();
                    manipList.Items.Clear();
                    condList.Items.Clear();

                    /*
                     * .IndexOf("|") returns -1 if "|" is not found, else it returns the zero based index
                     * of the first occurrence in which case we grab all chars from the start up to but
                     * not including the "|".
                    */
                    mnemonic = string.Format("{0}|{1}", (mnemonic.IndexOf("|") > -1 ? mnemonic.Substring(0, mnemonic.IndexOf("|")) : mnemonic), profileTag);
                }

                /*
                 * Checking for common block processing [Individual Dependents, Equifax Officers]:
                 * 
                 * The Ruby TK version was using regular expressions in the case statements 'when' lines,
                 * but the C# corollary, the switch statement, does not allow that. We were matching on
                 * a mnemonic value that STARTED with "IND|DEPENDENTS", [using Indiv Dependents as an
                 * example here] so we need to tweak things here and check for that scenario and reset
                 * mnemonic to "IND|DEPENDENTS" but keep a copy of the prior version so we can restore
                 * it back down below so we do not chop off any other mnemonic components from the other
                 * dependent associated list values.
                */
                match = Regex.Match(mnemonic, @"^(IND\|DEPENDENTS|EQU SUCCESSOR\|OFFICERS|EQU PREDECESSOR\|OFFICERS)");
                if (match.Success)
                {
                    mnemonic_copy = mnemonic;
                    mnemonic = match.Groups[1].ToString();
                }
                else
                {
                    /*
                     * Checking for Equifax State List processing:
                     * 
                     * We use a '##' placeholder for the state abbreviation rather than listing all of
                     * the possible mnemonic variations. We can retrieve the code from the snippet file
                     * using these, and we must adjust the mnemonic when posting to the 'Selection List'
                     * so they see the actual state abbreviation.
                    */
                    match = Regex.Match(mnemonic, @"^(EQU STATE LIST\|)[A-Z]{2}");
                    if (match.Success)
                    {
                        mnemonic = match.Groups[1].ToString() + "##";
                    }
                }

                //MessageBox.Show(String.Format("mnemonic: {0}", mnemonic));
                //Console.ReadLine();

                switch (mnemonic)
                {
                    /*#####################################################
		            #
		            # Individual Profile Section:
		            #
		            ######################################################*/

		            case ("IND|INDIVIDUAL"):

			            foreach (string str in indivIndArray) { subTagList.Items.Add(str); }
			            break;

		            case ("IND|SPOUSE"):

			            foreach (string str in indivIndArray) { subTagList.Items.Add(str); }

			            // Spouse has conditional dataflow ability.
			            foreach (string str in spouseCondArray) { condList.Items.Add(str); }
			
			            break;

		            case ("IND|INDIVIDUAL+SPOUSE"):

			            foreach (string str in indivIndSpArray) { subTagList.Items.Add(str); }
			            break;

		            case ("IND|BUSINESS"):

			            foreach (string str in indivBusArray) { subTagList.Items.Add(str); }
			            break;

		            case ("IND|BUSINESS/INDIVIDUAL"):

			            foreach (string str in indivBusIndArray) { subTagList.Items.Add(str); }
			            break;

		            case ("IND|BUSINESS/INDIVIDUAL+SPOUSE"):

			            foreach (string str in indivBusIndSpArray) { subTagList.Items.Add(str); }
			            break;

		            case ("IND|SOLE PROPRIETOR"):

			            foreach (string str in indivSoleArray) { subTagList.Items.Add(str); }

			            // Sole Proprietor has conditional dataflow ability.
			            foreach (string str in soleCondArray) { condList.Items.Add(str); }
			            break;

		            case ("IND|DEPENDENTS"):

                        if (tagListLocked)
                        {
                            if (profileTag == "<RETURN>")
                            {
                                LeavingCommonBlockProcessing(mnemonic);
                                DisplayMnemonicList("individual");
                                SetupProfile(null, profileLabel, tagList, subTagList, subTagCombList, manipList, condList, formBoxListBox, indivProfileArray);
                                mnemonic = "IND|";
                            }
                            else
                            {
                                numDependents = tagList.SelectedIndex + 1;

                                /* 
                                 * This is the only area where we need to restore mnemonic back to the copy
                                 * version.
                                */
                                mnemonic = mnemonic_copy;
                            }
                        }
                        else
                        {
                            EnteringCommonBlockProcessing(mnemonic);
                            DisplayMnemonicList("indivDep");
                        }
			            break;

		            case ("IND|FILING STATUS"):

			            foreach (string str in indivFileStatusArray) { subTagList.Items.Add(str); }
			            break;

		            case ("IND|DIRECT DEPOSIT"):

			            foreach (string str in indivDirectDepositArray) { subTagList.Items.Add(str); }
			            break;

                    /*#####################################################
		            #
		            # Entity Profile Section:
		            #
		            ######################################################*/

		            case ("ENT|GENERAL"):

                        foreach (string str in entGenArray) { subTagList.Items.Add(str); }
			            break;

		            case ("ENT|CONTACT"):

                        foreach (string str in entConArray) { subTagList.Items.Add(str); }
			            break;

		            case ("ENT|GENERAL+CONTACT"):

                        foreach (string str in entGenConArray) { subTagList.Items.Add(str); }
			            break;

		            case ("ENT|AGENT"):

                        foreach (string str in entAgentArray) { subTagList.Items.Add(str); }
			            break;

                    /*#####################################################
		            #
		            # Equifax Profile Section:
		            #
		            ######################################################*/

                    case ("EQU|GENERAL"):

                        /*
                         * Entering the General section of the Equifax profile.
                         * 
                         * Change mnemonic so we can properly handle Tag List selections since we are
                         * going to re-display the Tag List:
                        */
                        mnemonic = mnemonic.Replace('|', ' ') + '|';

                        /*
                         * Update profileLabel to indicate we are processing the 'General' section of
                         * the Equifax profile:
                        */
                        profileLabel.Text = profileLabel.Text + " - General";
                        tssLabelStatus.Text = profileLabel.Text;

                        tagList.Items.Clear();
                        foreach (string str in equGenArray) { tagList.Items.Add(str); }

                        // Display any Equifax-General selected mnemonics in the 'Selection List'.
                        DisplayMnemonicList("equGen");
                        break;

                    case ("EQU GENERAL|<RETURN>"):

                        // Leaving the General section of the Equifax profile.
                        tagList.Items.Clear();
                        foreach (string str in equProfileArray) { tagList.Items.Add(str); }

                        // Reset menmonic and profileLabel to reflect the Equifax initial state:
                        mnemonic = "EQU|";
                        profileLabel.Text = string.Format("{0} Profile", profile);
                        tssLabelStatus.Text = profileLabel.Text;

                        // Save the Equifax-General mnemonics in the 'Selection List'.
                        equGenMnemonicList = dfObjectList;

                        renumList.Items.Clear();
                        mnemonicListBox.Items.Clear();
                        dfObject = null;
                        break;

                    case ("EQU GENERAL|TAX ACCOUNTS REQUIRED"):

                        foreach (string str in equGenTaxAcctsReqArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU GENERAL|REASON FOR APPLICATION"):

                        foreach (string str in equReasonAppArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU|SUCCESSOR"):

                        /*
                         * Entering the Successor section of the Equifax profile.
                         * 
                         * Change mnemonic so we can properly handle Tag List selections since we are
                         * going to re-display the Tag List:
                        */
                        mnemonic = mnemonic.Replace('|', ' ') + '|';

                        /*
                         * Update profileLabel to indicate we are processing the 'Sucessor' section of
                         * the Equifax profile:
                        */
                        profileLabel.Text = profileLabel.Text + " - Successor";
                        tssLabelStatus.Text = profileLabel.Text;

                        tagList.Items.Clear();
                        foreach (string str in equSuccArray) { tagList.Items.Add(str); }

                        DisplayMnemonicList("equSucc");
                        break;

                    case ("EQU SUCCESSOR|<RETURN>"):

                        // Leaving the Successor section of the Equifax profile.
                        tagList.Items.Clear();
                        foreach (string str in equProfileArray) { tagList.Items.Add(str); }

                        // Reset menmonic and profileLabel to reflect the Equifax initial state:
                        mnemonic = "EQU|";
                        profileLabel.Text = string.Format("{0} Profile", profile);
                        tssLabelStatus.Text = profileLabel.Text;

                        // Make a copy of the 'Selection List'.
                        equSuccMnemonicList = dfObjectList;

                        renumList.Items.Clear();
                        mnemonicListBox.Items.Clear();
                        dfObject = null;
                        break;

                    case ("EQU PREDECESSOR|NAME"): goto case ("EQU SUCCESSOR|NAME");
                    case ("EQU SUCCESSOR|NAME"):

                        foreach (string str in equSuccPredNameArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|ADDRESS"):

                        foreach (string str in equSuccAddrNameArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|NAME+ADDRESS"):

                        foreach (string str in equSuccNameAddressArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU PREDECESSOR|FEIN"): goto case ("EQU SUCCESSOR|FEIN");
                    case ("EQU SUCCESSOR|FEIN"):

                        foreach (string str in feinArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|PAYROLL/TAX CONTACT"):

                        foreach (string str in equSuccConArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|ENTITY TYPE"):

                        foreach (string str in equSuccEntTypeArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|DATE STARTED"): goto case ("EQU SUCCESSOR|FISCAL YEAR END DATE");
                    case ("EQU SUCCESSOR|FIRST DAY OF EMPLOYMENT"): goto case ("EQU SUCCESSOR|FISCAL YEAR END DATE");
                    case ("EQU SUCCESSOR|FIRST PAYROLL DATE"): goto case ("EQU SUCCESSOR|FISCAL YEAR END DATE");
                    case ("EQU SUCCESSOR|20TH WEEK EMPLOYMENT DATE OF 1 OR MORE"): goto case ("EQU SUCCESSOR|FISCAL YEAR END DATE");
                    case ("EQU PREDECESSOR|DATE OF TRANSACTION"): goto case ("EQU SUCCESSOR|FISCAL YEAR END DATE");
                    case ("EQU SUCCESSOR|FISCAL YEAR END DATE"):

                        // Setting up the manipulation list here since nothing to display for the Sub-Tag list.
                        foreach (string str in dateManipArray) { manipList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|DATE STARTED+STATE OF INCORPORATION"):

                        foreach (string str in dateStartedStateOfIncArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|ACCOUNTING METHOD"):

                        foreach (string str in equSuccAccMethodArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|NATURE OF BUSINESS ACTIVITY"):

                        foreach (string str in equSuccNatBusActArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|501 (C)(3) EXEMPT"): goto case ("EQU SUCCESSOR|BREAK IN EMPLOYMENT");
                    case ("EQU PREDECESSOR|OPERATING AT TIME OF PURCHASE"): goto case ("EQU SUCCESSOR|BREAK IN EMPLOYMENT");
                    case ("EQU PREDECESSOR|EMPLOYMENT AGENCY"): goto case ("EQU SUCCESSOR|BREAK IN EMPLOYMENT");
                    case ("EQU SUCCESSOR|BREAK IN EMPLOYMENT"):

                        foreach (string str in yesNoChkBoxArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|SUBJECT TO FUTA"):

                        foreach (string str in equSuccFutaArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|PAYROLL SERVICE PROVIDER"):

                        foreach (string str in equSuccPaySvcProvArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|BANK ACCOUNT INFORMATION"):

                        foreach (string str in equSuccBankAccInfoArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|EMPLOYMENT IN OTHER STATES"):
                        foreach (string str in equSuccEmpOthStArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|TOTAL PAYROLL DATE OF $ OR MORE"):

                        foreach (string str in equSuccTotPayArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU PREDECESSOR|PARENT"): goto case ("EQU SUCCESSOR|PARENT");
                    case ("EQU SUCCESSOR|PARENT"):

                        foreach (string str in equSuccPredParentArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|OFFICERS"):

                        if (tagListLocked)
                        {
                            if (profileTag == "<RETURN>")
                            {
                                LeavingCommonBlockProcessing(mnemonic);
                                DisplayMnemonicList("equSucc");
                                SetupProfile(null, profileLabel, tagList, subTagList, subTagCombList, manipList, condList, formBoxListBox, equSuccArray);
                                mnemonic = "EQU SUCCESSOR|";
                            }
                            else
                            {
                                numEquSuccOfficers = tagList.SelectedIndex + 1;

                                // This is the only area where we need to restore mnemonic back to the copy version.
                                mnemonic = mnemonic_copy;
                            }
                        }
                        else
                        {
                            EnteringCommonBlockProcessing(mnemonic);
                            DisplayMnemonicList("equSuccOff");
                        }
                        break;

                    case ("EQU|PREDECESSOR"):

                        // Entering the Predecessor section of the Equifax profile.

                        // Change mnemonic so we can properly handle Tag List selections since we are going to re-display the Tag List:
                        mnemonic = mnemonic.Replace('|', ' ') + '|';

                        // Update profileLabel to indicate we are processing the 'Predecessor' section of the Equifax profile:
                        profileLabel.Text = profileLabel.Text + " - Predecessor";
                        tssLabelStatus.Text = profileLabel.Text;

                        tagList.Items.Clear();
                        foreach (string str in equPredArray) { tagList.Items.Add(str); }

                        DisplayMnemonicList("equPred");
                        break;

                    case ("EQU PREDECESSOR|<RETURN>"):

                        // Leaving the Predecessor section of the Equifax profile.

                        tagList.Items.Clear();
                        foreach (string str in equProfileArray) { tagList.Items.Add(str); }

                        // Reset menmonic and profileLabel to reflect the Equifax initial state:
                        mnemonic = "EQU|";
                        profileLabel.Text = string.Format("{0} Profile", profile);
                        tssLabelStatus.Text = profileLabel.Text;

                        // Make a copy of the 'Selection List' and initialize to the default Equifax profile list.
                        equPredMnemonicList = dfObjectList;

                        renumList.Items.Clear();
                        mnemonicListBox.Items.Clear();
                        dfObject = null;
                        break;

                    case ("EQU PREDECESSOR|CORPORATE ADDRESS"):

                        // Not using the subTagList to stay consistant with other profiles.
                        foreach (string str in addressArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU PREDECESSOR|CONTACT"):

                        foreach (string str in equPredConArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU PREDECESSOR|OFFICERS"):

                        if (tagListLocked)
                        {
                            if (profileTag == "<RETURN>")
                            {
                                LeavingCommonBlockProcessing(mnemonic);
                                DisplayMnemonicList("equPred");
                                SetupProfile(null, profileLabel, tagList, subTagList, subTagCombList, manipList, condList, formBoxListBox, equPredArray);
                                mnemonic = "EQU PREDECESSOR|";
                            }
                            else
                            {
                                numEquPredOfficers = tagList.SelectedIndex + 1;

                                // This is the only area where we need to restore mnemonic back to the copy version.
                                mnemonic = mnemonic_copy;
                            }
                        }
                        else
                        {
                            EnteringCommonBlockProcessing(mnemonic);
                            DisplayMnemonicList("equPredOff");
                        }
                        break;

                    case ("EQU PREDECESSOR|TRANSACTION TYPE"):

                        foreach (string str in equPredTranTypeArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU PREDECESSOR|ALL/PART EMPLOYMENT IN EACH STATE"):

                        foreach (string str in equPredAllPartArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU PREDECESSOR|COMMONLY OWNED"):

                        foreach (string str in equPredCommOwnedArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU PREDECESSOR|ASSETS TRANSFERRED/ACQUIRED"):

                        foreach (string str in equPredAssTranAcqArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU PREDECESSOR|CONTINUE TO REPORT WAGES"):

                        foreach (string str in equPredContRptWagesArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU|STATE LIST"):

                        // Entering the State List section of the Equifax profile.

                        // Change mnemonic so we can properly handle Tag List selections since we are going to re-display the Tag List:
                        mnemonic = mnemonic.Replace('|', ' ') + '|';

                        // Update profileLabel to indicate we are processing the 'State List' section of the Equifax profile:
                        profileLabel.Text = profileLabel.Text + " - State List";
                        tssLabelStatus.Text = profileLabel.Text;

                        tagList.Items.Clear();
                        foreach (string str in equStateArray) { tagList.Items.Add(str); }

                        // Display any Equifax-State List selected mnemonics in the 'Selection List'.
                        DisplayMnemonicList("equState");
                        break;

                    case ("EQU STATE LIST|<RETURN>"):

                        // Leaving the State List section of the Equifax profile.
                        tagList.Items.Clear();
                        foreach (string str in equProfileArray) { tagList.Items.Add(str); }

                        // Reset menmonic and profileLabel to reflect the Equifax initial state:
                        mnemonic = "EQU|";
                        profileLabel.Text = string.Format("{0} Profile", profile);
                        tssLabelStatus.Text = profileLabel.Text;

                        // Save the Equifax-State List mnemonics in the 'Selection List'.
                        equStateMnemonicList = dfObjectList;

                        renumList.Items.Clear();
                        mnemonicListBox.Items.Clear();
                        dfObject = null;
                        break;

                    case ("EQU STATE LIST|##"):

                        foreach (string str in equStateSubTagArray) { subTagList.Items.Add(str); }
                        break;

                    case ("EQU PREDECESSOR|PHONE"): goto case ("EQU SUCCESSOR|PHONE");
                    case ("EQU SUCCESSOR|PHONE"):

                        // Setting up the manipulation list here since nothing to display for the Sub-Tag list.
                        foreach (string str in phoneManipArray) { manipList.Items.Add(str); }
                        break;

                    /*#####################################################
		            #
		            # Fiduciary Profile Section:
		            #
		            ######################################################*/

		            case ("FID|FIDUCIARY"):

                        foreach (string str in fidFidArray) { subTagList.Items.Add(str); }
                        break;

		            case ("FID|ESTATE/TRUST"):

                        foreach (string str in fidEsTrArray) { subTagList.Items.Add(str); }
                        break;

		            case ("FID|DECEDENT/DONOR"):

                        foreach (string str in fidDecDnArray) { subTagList.Items.Add(str); }
                        break;

		            case ("FID|ESTATE/TRUST+DECEDENT/DONOR"):

                        foreach (string str in fidEsTrDecDnArray) { subTagList.Items.Add(str); }
                        break;

                    /*#####################################################
		            #
		            # Preparer Profile Section:
		            #
		            ######################################################*/

		            case ("PRE|NAME"):

                        foreach (string str in preNameArray) { subTagList.Items.Add(str); }
                        break;

		            case ("PRE|ADDRESS"):

			            // Setting up the combinations list here since there is nothing to display for the Sub-Tag list.
                        foreach (string str in addrArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|NAME+ADDRESS"):

                        foreach (string str in preNameAddressArray) { subTagList.Items.Add(str); }
                        break;

		            case ("PRE|SELF EMPLOYED"):

			            // Setting up the manipulation list here since nothing to display for the Sub-Tag list.
                        foreach (string str in preSempManipArray) { manipList.Items.Add(str); }
                        break;

		            case ("PRE|SELF EMPLOYED+NAME"):

                        foreach (string str in preSempNameArray) { subTagList.Items.Add(str); }
                        break;

		            case ("PRE|SELF EMPLOYED+NAME+ADDRESS"):

                        foreach (string str in preSempNameAddrArray) { subTagList.Items.Add(str); }
                        break;

		            case ("PRE|ID NUMBER"):

                        foreach (string str in preIdArray) { subTagList.Items.Add(str); }
                        break;

		            case ("PRE|NAME+ID NUMBER"):

                        foreach (string str in preNameIdArray) { subTagList.Items.Add(str); }
                        break;

		            case ("PRE|PHONE"):

			            // Setting up the combinations list here since there is nothing to display for the Sub-Tag list.
                        foreach (string str in prePhoneCombArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|NAME+PHONE"):

                        foreach (string str in preNamePhoneArray) { subTagList.Items.Add(str); }
                        break;

		            case ("PRE|PHONE+FAX"):

			            // Setting up the combinations list here since there is nothing to display for the Sub-Tag list.
                        foreach (string str in prePhoneFaxCombArray) { subTagCombList.Items.Add(str); }
                        break;
                }
            }
        }

        private void subTagList_SelectedIndexChanged(object sender, EventArgs e)
        {
            // Return if nothing is selected in the list.
            if (subTagList.SelectedIndex != -1)
            {
                string subTag = ((String)subTagList.SelectedItem).ToUpper();

                // Clear the other lists.
                subTagCombList.Items.Clear();
                manipList.Items.Clear();

                // .IndexOf(":") returns -1 if ":" is not found, else it returns the zero based index of the first occurrence in which case
                // we grab all chars from the start up to but not including the ":".
                mnemonic = string.Format("{0}:{1}", (mnemonic.IndexOf(":") > -1 ? mnemonic.Substring(0, mnemonic.IndexOf(":")) : mnemonic), subTag);

                switch (mnemonic)
                {
                    // C# switch statements are not as generous as Ruby, we can not use regular expressions here, so we
                    // do a lot of fall throughs here regarding case's.

                    /*#####################################################
		            #
		            # Individual Profile Section:
		            #
		            ######################################################*/

                    case ("IND|INDIVIDUAL:NAME"): goto case ("IND|SPOUSE:NAME");
                    case ("IND|SPOUSE:NAME"):

                        foreach (string str in indivIndNameArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("IND|INDIVIDUAL:ADDRESS"): goto case ("IND|SPOUSE:ADDRESS");
                    case ("IND|SPOUSE:ADDRESS"):

                        foreach (string str in indivIndAddressArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("IND|INDIVIDUAL:NAME+ADDRESS"): goto case ("IND|SPOUSE:NAME+ADDRESS");
                    case ("IND|SPOUSE:NAME+ADDRESS"):

                        foreach (string str in indivIndNameAddressArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("IND|INDIVIDUAL:PHONE"): goto case ("IND|SPOUSE:PHONE");
                    case ("IND|SPOUSE:PHONE"):

                        foreach (string str in indivIndPhoneArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("IND|INDIVIDUAL:NAME+PHONE"): goto case ("IND|SPOUSE:NAME+PHONE");
                    case ("IND|SPOUSE:NAME+PHONE"):

                        foreach (string str in indivIndNamePhoneArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("IND|INDIVIDUAL:PHONE+FAX"): goto case ("IND|SPOUSE:PHONE+FAX");
                    case ("IND|SPOUSE:PHONE+FAX"):

                        foreach (string str in indivIndPhoneFaxArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("IND|INDIVIDUAL:SSN"):

                        foreach (string str in indSsnArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("IND|SPOUSE:SSN"): goto case ("IND|DEPENDENTS:SSN");
                    case ("IND|DEPENDENTS:SSN"):

                        foreach (string str in ssnArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("IND|INDIVIDUAL:BLIND"): goto case ("IND|SPOUSE:BLIND");
                    case ("IND|SPOUSE:BLIND"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in blindManipArray) { manipList.Items.Add(str); }
                        break;

                    case ("IND|INDIVIDUAL:65+"): goto case ("IND|SPOUSE:65+");
                    case ("IND|SPOUSE:65+"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in sixtyFivePlusManipArray) { manipList.Items.Add(str); }
                        break;

		            case ("IND|INDIVIDUAL+SPOUSE:NAME"):

                        foreach (string str in indivIndSpNameArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("IND|INDIVIDUAL+SPOUSE:SSN"):

                        foreach (string str in indivIndSpSsnArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("IND|BUSINESS:ADDRESS"):

                        foreach (string str in addressArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("IND|BUSINESS:NAME+ADDRESS"):

                        foreach (string str in indivBusNameAddressArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("IND|BUSINESS:PHONE"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in phoneManipArray) { manipList.Items.Add(str); }
                        break;

		            case ("IND|BUSINESS:FEIN"):

                        foreach (string str in feinArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("IND|BUSINESS/INDIVIDUAL:PHONE"):

                        foreach (string str in indivBusIndPhoneArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("IND|BUSINESS/INDIVIDUAL+SPOUSE:NAME"):

                        foreach (string str in indivBusIndSpNameArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("IND|SOLE PROPRIETOR:NAME"):

                        foreach (string str in indivSoleNameArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("IND|SOLE PROPRIETOR:ADDRESS"):

                        foreach (string str in indivSoleAddressArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("IND|SOLE PROPRIETOR:PHONE"):

                        foreach (string str in indivSolePhoneArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("IND|SOLE PROPRIETOR:NAME+PHONE"):

                        foreach (string str in indivSoleNamePhoneArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("IND|SOLE PROPRIETOR:PHONE+FAX"):

                        foreach (string str in indivSolePhoneFaxArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("IND|SOLE PROPRIETOR:FEIN / SSN"):

                        foreach (string str in feinSsnArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("IND|DEPENDENTS:NAME"):

                        foreach (string str in indivDepNameArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("IND|DEPENDENTS:DOB"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in dateManipArray) { manipList.Items.Add(str); }
                        break;

		            case ("IND|FILING STATUS:SINGLE"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in singleManipArray) { manipList.Items.Add(str); }
                        break;

		            case ("IND|FILING STATUS:MARRIED FILING JOINT"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in jointManipArray) { manipList.Items.Add(str); }
                        break;

		            case ("IND|FILING STATUS:MARRIED FILING SEPARATE"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in separateManipArray) { manipList.Items.Add(str); }
                        break;

		            case ("IND|FILING STATUS:MARRIED FILING JOINT OR SEPARATE"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in jointSeparateManipArray) { manipList.Items.Add(str); }
                        break;

		            case ("IND|FILING STATUS:HEAD OF HOUSEHOLD"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in hohManipArray) { manipList.Items.Add(str); }
                        break;

		            case ("IND|FILING STATUS:WIDOW(ER)"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in widowManipArray) { manipList.Items.Add(str); }
                        break;

		            case ("IND|DIRECT DEPOSIT:CHECKING ACCOUNT"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in ddCkManipArray) { manipList.Items.Add(str); }
                        break;

		            case ("IND|DIRECT DEPOSIT:SAVINGS ACCOUNT"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in ddSvManipArray) { manipList.Items.Add(str); }
                        break;

                    /*#####################################################
		            #
		            # Entity Profile Section:
		            #
		            ######################################################*/

		            case ("ENT|GENERAL:ADDRESS"):

                        foreach (string str in addressArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("ENT|GENERAL:NAME+ADDRESS"):

                        foreach (string str in entGenNameAddressArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("ENT|GENERAL:PHONE"): goto case ("ENT|AGENT:PHONE");
                    case ("ENT|CONTACT:PHONE"): goto case ("ENT|AGENT:PHONE");
                    case ("ENT|AGENT:PHONE"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in phoneManipArray) { manipList.Items.Add(str); }
                        break;

		            case ("ENT|GENERAL:FEIN"):

                        foreach (string str in entFeinArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("ENT|GENERAL:DATE STARTED"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in dateManipArray) { manipList.Items.Add(str); }
                        break;

		            case ("ENT|GENERAL:DATE STARTED+STATE OF INCORPORATION "):

                        foreach (string str in dateStartedStateOfIncArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("ENT|GENERAL+CONTACT:NAME"):

                        foreach (string str in entGenConNameArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("ENT|GENERAL+CONTACT:PHONE"):

                        foreach (string str in entGenConPhoneArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("ENT|GENERAL+CONTACT:NAME+PHONE"):

                        foreach (string str in entGenConNamePhoneArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("ENT|AGENT:NAME"):

                        foreach (string str in entAgentNameArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("ENT|AGENT:ADDRESS"):

                        foreach (string str in addrArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("ENT|AGENT:NAME+ADDRESS"):

                        foreach (string str in entAgentNameAddressArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("ENT|AGENT:NAME+PHONE"):

                        foreach (string str in entAgentNamePhoneArray) { subTagCombList.Items.Add(str); }
                        break;

                    /*#####################################################
		            #
		            # Equifax Profile Section:
		            #
		            ######################################################*/

                    case ("EQU GENERAL|TAX ACCOUNTS REQUIRED:UNEMPLOYMENT TAX"): goto case ("EQU GENERAL|TAX ACCOUNTS REQUIRED:EMPLOYMENT TAX CONSULTING");
                    case ("EQU GENERAL|TAX ACCOUNTS REQUIRED:LOCAL WITHHOLDING TAX"): goto case ("EQU GENERAL|TAX ACCOUNTS REQUIRED:EMPLOYMENT TAX CONSULTING");
                    case ("EQU GENERAL|TAX ACCOUNTS REQUIRED:WITHHOLDING TAX"): goto case ("EQU GENERAL|TAX ACCOUNTS REQUIRED:EMPLOYMENT TAX CONSULTING");
                    case ("EQU GENERAL|TAX ACCOUNTS REQUIRED:EMPLOYMENT TAX CONSULTING"):

                        foreach (string str in yesNoChkBoxArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU GENERAL|REASON FOR APPLICATION:OTHER"): goto case ("EQU SUCCESSOR|ENTITY TYPE:OTHER");
                    case ("EQU SUCCESSOR|NATURE OF BUSINESS ACTIVITY:OTHER"): goto case ("EQU SUCCESSOR|ENTITY TYPE:OTHER");
                    case ("EQU PREDECESSOR|TRANSACTION TYPE:OTHER"): goto case ("EQU SUCCESSOR|ENTITY TYPE:OTHER");
                    case ("EQU SUCCESSOR|ENTITY TYPE:OTHER"):

                        foreach (string str in equOtherArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|ADDRESS:CORPORATE"): goto case ("EQU SUCCESSOR|ADDRESS:MAILING");
                    case ("EQU PREDECESSOR|CONTACT:ADDRESS"): goto case ("EQU SUCCESSOR|ADDRESS:MAILING");
                    case ("EQU SUCCESSOR|ADDRESS:MAILING"):

                        foreach (string str in addressArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU PREDECESSOR|CONTACT:NAME+ADDRESS"):

                        foreach (string str in equPredConNameAddrArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|NAME+ADDRESS:LEGAL NAME+CORPORATE ADDRESS"): goto case ("EQU SUCCESSOR|NAME+ADDRESS:LEGAL NAME+MAILING ADDRESS");
                    case ("EQU SUCCESSOR|NAME+ADDRESS:LEGAL NAME+MAILING ADDRESS"):

                        foreach (string str in equSuccLegalNameAddressArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU PREDECESSOR|CONTACT:PHONE"): goto case ("EQU SUCCESSOR|PAYROLL SERVICE PROVIDER:PHONE");
                    case ("EQU SUCCESSOR|PAYROLL/TAX CONTACT:PHONE"): goto case ("EQU SUCCESSOR|PAYROLL SERVICE PROVIDER:PHONE");
                    case ("EQU SUCCESSOR|BANK ACCOUNT INFORMATION:PHONE"): goto case ("EQU SUCCESSOR|PAYROLL SERVICE PROVIDER:PHONE");
                    case ("EQU SUCCESSOR|PAYROLL SERVICE PROVIDER:PHONE"):

                        // Setting up the manipulation list here since nothing to display for the Sub-Tag Combination list.
                        foreach (string str in phoneManipArray) { manipList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|BANK ACCOUNT INFORMATION:ADDRESS"): goto case ("EQU SUCCESSOR|PAYROLL SERVICE PROVIDER:ADDRESS");
                    case ("EQU PREDECESSOR|OFFICERS:ADDRESS"): goto case ("EQU SUCCESSOR|PAYROLL SERVICE PROVIDER:ADDRESS");
                    case ("EQU SUCCESSOR|OFFICERS:ADDRESS"): goto case ("EQU SUCCESSOR|PAYROLL SERVICE PROVIDER:ADDRESS");
                    case ("EQU SUCCESSOR|PAYROLL SERVICE PROVIDER:ADDRESS"):

                        foreach (string str in addrArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|PAYROLL SERVICE PROVIDER:PROVIDER NAME+ADDRESS"): goto case ("EQU PREDECESSOR|OFFICERS:NAME+ADDRESS");
                    case ("EQU SUCCESSOR|BANK ACCOUNT INFORMATION:BANK NAME+ADDRESS"): goto case ("EQU PREDECESSOR|OFFICERS:NAME+ADDRESS");
                    case ("EQU SUCCESSOR|OFFICERS:NAME+ADDRESS"): goto case ("EQU PREDECESSOR|OFFICERS:NAME+ADDRESS");
                    case ("EQU PREDECESSOR|OFFICERS:NAME+ADDRESS"):

                        foreach (string str in equNameAddrArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|PAYROLL SERVICE PROVIDER:PROVIDER NAME+ADDRESS+PHONE"): goto case ("EQU SUCCESSOR|BANK ACCOUNT INFORMATION:BANK NAME+ADDRESS+PHONE");
                    case ("EQU SUCCESSOR|BANK ACCOUNT INFORMATION:BANK NAME+ADDRESS+PHONE"):

                        foreach (string str in equNameAddrPhoneArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|EMPLOYMENT IN OTHER STATES:YES"):
                        foreach (string str in equSuccEmpOthStYesArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|OFFICERS:NAME+TITLE+ADDRESS"): goto case ("EQU PREDECESSOR|OFFICERS:NAME+TITLE+ADDRESS");
                    case ("EQU PREDECESSOR|OFFICERS:NAME+TITLE+ADDRESS"):

                        foreach (string str in equNameTitleAddrArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|ENTITY TYPE:CORPORATION (PUBLICLY TRADED)"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU SUCCESSOR|ENTITY TYPE:CORPORATION (PRIVATELY HELD)"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU SUCCESSOR|ENTITY TYPE:PARTNERSHIP"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU SUCCESSOR|ENTITY TYPE:LIMITED PARTNERSHIP"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU SUCCESSOR|ENTITY TYPE:S CORPORATION"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU SUCCESSOR|ENTITY TYPE:LIMITED LIABILITY COMPANY"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU SUCCESSOR|ENTITY TYPE:TAXED AS CORPORATION"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU SUCCESSOR|ENTITY TYPE:TAXED AS PARTNERSHIP"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU SUCCESSOR|NATURE OF BUSINESS ACTIVITY:RETAIL"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU SUCCESSOR|NATURE OF BUSINESS ACTIVITY:WHOLESALE"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU SUCCESSOR|NATURE OF BUSINESS ACTIVITY:MANUFACTURING"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU SUCCESSOR|NATURE OF BUSINESS ACTIVITY:SERVICE"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU SUCCESSOR|NATURE OF BUSINESS ACTIVITY:STAFFING/TEMP/EMPLY AGT SERVICES"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU SUCCESSOR|NATURE OF BUSINESS ACTIVITY:PROFESSIONAL EMPLOYER ORGANIZATION"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU SUCCESSOR|ACCOUNTING METHOD:CASH"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU SUCCESSOR|ACCOUNTING METHOD:ACCRUAL"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|TRANSACTION TYPE:ASSET PURCHASE"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|TRANSACTION TYPE:STOCK PURCHASE"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|TRANSACTION TYPE:COMPANY REORGANIZATION"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|TRANSACTION TYPE:SALUTORY MERGER"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|TRANSACTION TYPE:EMPLOYEE REASSIGNMENT ONLY"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|TRANSACTION TYPE:BANKRUPTCY ACQUISITION"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|TRANSACTION TYPE:ENTITY CONVERSION"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|ALL/PART EMPLOYMENT IN EACH STATE:ALL"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|ALL/PART EMPLOYMENT IN EACH STATE:PARTIAL"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|ASSETS TRANSFERRED/ACQUIRED:ACCOUNTS RECEIVABLE"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|ASSETS TRANSFERRED/ACQUIRED:MACHINERY OR EQUIPMENT"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|ASSETS TRANSFERRED/ACQUIRED:FURNITURE OR FIXTURES"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|ASSETS TRANSFERRED/ACQUIRED:GOODWILL"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|ASSETS TRANSFERRED/ACQUIRED:BUILDINGS"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|ASSETS TRANSFERRED/ACQUIRED:STOCK"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|ASSETS TRANSFERRED/ACQUIRED:OTHER REAL PROPERTY OR LEASES"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|ASSETS TRANSFERRED/ACQUIRED:CUSTOMER LISTS"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|ASSETS TRANSFERRED/ACQUIRED:EMPLOYEES"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU PREDECESSOR|ASSETS TRANSFERRED/ACQUIRED:INVENTORIES"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU GENERAL|REASON FOR APPLICATION:OPENING A NEWLY FORMED BUSINESS"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU GENERAL|REASON FOR APPLICATION:EXISTING ENTITY - NEW IN STATE(S)"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU GENERAL|REASON FOR APPLICATION:DID YOU ACQUIRE ALL/PART OF AN EXISTING BUSINESS"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU GENERAL|REASON FOR APPLICATION:REOPENING A CLOSED/INACTIVE BUSINESS"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU GENERAL|REASON FOR APPLICATION:CHANGING OWNERSHIP TYPE/ENTITY CHANGE"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU GENERAL|REASON FOR APPLICATION:NAME CHANGE"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU GENERAL|REASON FOR APPLICATION:ACCOUNT CLOSURES"): goto case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP");
                    case ("EQU SUCCESSOR|ENTITY TYPE:SOLE PROPRIETORSHIP"):

                        foreach (string str in chkBoxArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|ENTITY TYPE:DISREGARDED ENTITY"):

                        foreach (string str in equDisEntArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|SUBJECT TO FUTA:FIRST LIABILITY YEAR"): goto case ("EQU SUCCESSOR|TOTAL PAYROLL DATE OF $ OR MORE:YEAR");
                    case ("EQU SUCCESSOR|SUBJECT TO FUTA:TWO BOXES: CHECK ['X'] YES BOX; FIRST LIABILITY YEAR"): goto case ("EQU SUCCESSOR|TOTAL PAYROLL DATE OF $ OR MORE:YEAR");
                    case ("EQU SUCCESSOR|TOTAL PAYROLL DATE OF $ OR MORE:QUARTER+YEAR"): goto case ("EQU SUCCESSOR|TOTAL PAYROLL DATE OF $ OR MORE:YEAR");
                    case ("EQU SUCCESSOR|TOTAL PAYROLL DATE OF $ OR MORE:YEAR"):

                        // Setting up the manipulation list here since nothing to display for the Sub-Tag Combination list.
                        foreach (string str in yearManipArray) { manipList.Items.Add(str); }
                        break;

                    case ("EQU PREDECESSOR|OFFICERS:SSN"): goto case ("EQU SUCCESSOR|OFFICERS:SSN");
                    case ("EQU SUCCESSOR|OFFICERS:SSN"):
                        foreach (string str in ssnArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU SUCCESSOR|PARENT:FEIN"): goto case ("EQU SUCCESSOR|OFFICERS:FEIN");
                    case ("EQU PREDECESSOR|PARENT:FEIN"): goto case ("EQU SUCCESSOR|OFFICERS:FEIN");
                    case ("EQU PREDECESSOR|OFFICERS:FEIN"): goto case ("EQU SUCCESSOR|OFFICERS:FEIN");
                    case ("EQU SUCCESSOR|OFFICERS:FEIN"):
                        foreach (string str in feinArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU PREDECESSOR|OFFICERS:SSN / FEIN"): goto case ("EQU SUCCESSOR|OFFICERS:SSN / FEIN");
                    case ("EQU SUCCESSOR|OFFICERS:SSN / FEIN"):
                        foreach (string str in feinSsnArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU PREDECESSOR|CONTINUE TO REPORT WAGES:LAST DATE OF EMPLOYMENT"): goto case ("EQU SUCCESSOR|OFFICERS:DATE OF APPOINTMENT");
                    case ("EQU PREDECESSOR|CONTINUE TO REPORT WAGES:DATE OF LAST PAYROLL"): goto case ("EQU SUCCESSOR|OFFICERS:DATE OF APPOINTMENT");
                    case ("EQU PREDECESSOR|CONTINUE TO REPORT WAGES:CHECK ['X'] NO BOX; LAST DATE OF EMPLOYMENT"): goto case ("EQU SUCCESSOR|OFFICERS:DATE OF APPOINTMENT");
                    case ("EQU PREDECESSOR|CONTINUE TO REPORT WAGES:CHECK ['X'] NO BOX; DATE OF LAST PAYROLL"): goto case ("EQU SUCCESSOR|OFFICERS:DATE OF APPOINTMENT");
                    // BOTH dates get the same manipulation for this one - only way to do it if offering date validation.
                    case ("EQU PREDECESSOR|CONTINUE TO REPORT WAGES:CHECK ['X'] NO BOX; LAST DATE OF EMPLOYMENT; DATE OF LAST PAYROLL"): goto case ("EQU SUCCESSOR|OFFICERS:DATE OF APPOINTMENT");
                    case ("EQU PREDECESSOR|OFFICERS:DOB"): goto case ("EQU SUCCESSOR|OFFICERS:DATE OF APPOINTMENT");
                    case ("EQU PREDECESSOR|OFFICERS:DATE OF APPOINTMENT"): goto case ("EQU SUCCESSOR|OFFICERS:DATE OF APPOINTMENT");
                    case ("EQU SUCCESSOR|OFFICERS:DOB"): goto case ("EQU SUCCESSOR|OFFICERS:DATE OF APPOINTMENT");
                    case ("EQU SUCCESSOR|OFFICERS:DATE OF APPOINTMENT"):

                        // Setting up the manipulation list here since nothing to display for the Sub-Tag Combination list.
                        foreach (string str in dateManipArray) { manipList.Items.Add(str); }
                        break;

                    case ("EQU STATE LIST|##:ACCOUNT NUMBERS"):

                        foreach (string str in equStateAcctNumArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU STATE LIST|##:NUMBER OF EMPLOYEES"):

                        foreach (string str in equStateNumEmpArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("EQU STATE LIST|##:WITHHOLDING TAX"):

                        foreach (string str in equStateWithTaxArray) { subTagCombList.Items.Add(str); }
                        break;

                    /*#####################################################
		            #
		            # Fiduciary Profile Section:
		            #
		            ######################################################*/

                    case ("FID|FIDUCIARY:NAME"): goto case ("FID|DECEDENT/DONOR:NAME");
		            case ("FID|DECEDENT/DONOR:NAME"):

                        foreach (string str in fidNameArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("FID|FIDUCIARY:ADDRESS"): goto case ("FID|DECEDENT/DONOR:ADDRESS");
		            case ("FID|DECEDENT/DONOR:ADDRESS"):

                        foreach (string str in addrArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("FID|FIDUCIARY:NAME+ADDRESS"): goto case ("FID|DECEDENT/DONOR:NAME+ADDRESS");
		            case ("FID|DECEDENT/DONOR:NAME+ADDRESS"):

                        foreach (string str in fidNameAddressArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("FID|FIDUCIARY:NAME+TITLE"):

                        foreach (string str in fidFidNameTitleArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("FID|FIDUCIARY:PHONE"): goto case ("FID|DECEDENT/DONOR:PHONE");
                    case ("FID|DECEDENT/DONOR:PHONE"):

			            // Setting up the manipulation list here since nothing to display for the Sub-Tag Combination list.
                        foreach (string str in phoneManipArray) { manipList.Items.Add(str); }
                        break;

                    case ("FID|FIDUCIARY:NAME+PHONE"): goto case ("FID|DECEDENT/DONOR:NAME+PHONE");
                    case ("FID|DECEDENT/DONOR:NAME+PHONE"):

                        foreach (string str in fidFidNamePhoneArray) { subTagCombList.Items.Add(str); }
                        break;

                    case ("FID|FIDUCIARY:SSN"): goto case ("FID|DECEDENT/DONOR:SSN");
                    case ("FID|ESTATE/TRUST:SSN"): goto case ("FID|DECEDENT/DONOR:SSN");
                    case ("FID|DECEDENT/DONOR:SSN"):

                        foreach (string str in ssnArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("FID|ESTATE/TRUST:FEIN"):

                        foreach (string str in feinArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("FID|ESTATE/TRUST:FEIN / SSN"):

                        foreach (string str in feinSsnArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("FID|ESTATE/TRUST:CREATION DATE"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in dateManipArray) { manipList.Items.Add(str); }
                        break;

		            case ("FID|ESTATE/TRUST:TYPE"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in fidEsTrTypeManipArray) { manipList.Items.Add(str); }
                        break;

                    case ("FID|DECEDENT/DONOR:DOB"): goto case ("FID|DECEDENT/DONOR:DOD");
                    case ("FID|DECEDENT/DONOR:DOD"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in dateManipArray) { manipList.Items.Add(str); }
                        break;

		            case ("FID|ESTATE/TRUST+DECEDENT/DONOR:DOD / CREATION DATE"):

			            // Setting up the manipulation list here since there is nothing to display for the Sub-Tag
			            // Combination list.
                        foreach (string str in dateManipArray) { manipList.Items.Add(str); }
                        break;

                    /*#####################################################
		            #
		            # Preparer Profile Section:
		            #
		            ######################################################*/

		            case ("PRE|NAME:INDIVIDUAL"):

                        foreach (string str in preIndNameCombArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|NAME:FIRM / INDIVIDUAL"):

                        foreach (string str in preFirmIndNameCombArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|NAME+ADDRESS:[FIRM / INDIVIDUAL]+ADDRESS"):

                        foreach (string str in preFirmIndNameAddrCombArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|SELF EMPLOYED+NAME:NAME BOX"):

                        foreach (string str in preSempNameBoxArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|SELF EMPLOYED+NAME:NAME BOX; CHECK BOX: IF SELF EMPLOYED"):

                        foreach (string str in preSempNameChkBoxArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|SELF EMPLOYED+NAME:NAME BOX[PREP.]; NAME BOX[FIRM / PREP. IF SELF EMP.]; CHECK BOX: IF SELF EMP."):

                        foreach (string str in preSempNamesChkBoxArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|SELF EMPLOYED+NAME+ADDRESS:NAME+ADDRESS BOX"):

                        foreach (string str in preSempNameAddrBoxArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|SELF EMPLOYED+NAME+ADDRESS:NAME+ADDRESS BOX; CHECK BOX: IF SELF EMPLOYED"):

                        foreach (string str in preSempNameAddrChkBoxArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|ID NUMBER:FEIN"):

                        foreach (string str in feinArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|ID NUMBER:SSN"):

                        foreach (string str in ssnArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|ID NUMBER:PTIN / FEIN / SSN"):

                        foreach (string str in ptinFeinSsnArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|ID NUMBER:PTIN / FEIN"):

                        foreach (string str in ptinFeinArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|ID NUMBER:PTIN / SSN"):

                        foreach (string str in ptinSsnArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|ID NUMBER:FEIN / PTIN"):

                        foreach (string str in feinPtinArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|ID NUMBER:FEIN / SSN"):

                        foreach (string str in feinSsnArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|NAME+ID NUMBER:[FIRM / INDIVIDUAL]+ID NUMBER"):

                        foreach (string str in preNameIdCombArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|NAME+PHONE:INDIVIDUAL+PHONE"):

                        foreach (string str in preNamePhoneIndArray) { subTagCombList.Items.Add(str); }
                        break;

		            case ("PRE|NAME+PHONE:[FIRM / INDIVIDUAL]+PHONE"):

                        foreach (string str in preNamePhoneFirmIndArray) { subTagCombList.Items.Add(str); }
                        break;
                }
            }
        }

        private void subTagCombList_SelectedIndexChanged(object sender, EventArgs e)
        {
            // Return if nothing is selected in the list.
            if (subTagCombList.SelectedIndex != -1)
            {
                string subTagComb = ((String)subTagCombList.SelectedItem).ToUpper();

	            // Update the mnemonic based on list selection unless the selection is "<None>". Also clear the other lists if applicable.
                // NOTE: I confused myself here with my own reference to "NONE" [Ruby version] as this list has no "<None>" items, but will
                // do this part like the Ruby verion anyways, could be a subtle 'gotcha' lurking around here...
                if (subTagComb != "<NONE>")
                {

                    // .IndexOf("::") returns -1 if "::" is not found, else it returns the zero based index of the first occurrence in which case
                    // we grab all chars from the start up to but not including the "::".
                    mnemonic = string.Format("{0}::{1}", (mnemonic.IndexOf("::") > -1 ? mnemonic.Substring(0, mnemonic.IndexOf("::")) : mnemonic), subTagComb);

                    manipList.Items.Clear();
                }

                /*
                Going to use regular expressions for this list since there are many permutations per list selection.
                We must use if..else if..statements for this since C# does not allow regular expressions for switching.
                Not using the ^ and $ anchors for this, we control the mnemonic, so it always starts with known characters,
                we want to find in most cases, a pattern which will not be the whole mnemonic, but match up to a point.
                */

                /*#####################################################
                #
                # Individual Profile Section:
                #
                ######################################################*/

                if (Regex.IsMatch(mnemonic, @"IND\|(INDIVIDUAL|SPOUSE|BUSINESS):ADDRESS::ZIP"))
                {
                    foreach (string str in zipManipArray) { manipList.Items.Add(str); }
                }
                // NOTE: We have '::' in the regular expression so that we do not erroneously match with
		        // @"IND\|...:PHONE+FAX", as we are not using the ^ and $ characters to start and stop the match.
                else if (Regex.IsMatch(mnemonic, @"IND\|(INDIVIDUAL|SPOUSE|BUSINESS\/INDIVIDUAL|SOLE PROPRIETOR):PHONE::"))
                {
                    foreach (string str in phoneManipArray) { manipList.Items.Add(str); }
                }
                else if (Regex.IsMatch(mnemonic, @"IND\|(INDIVIDUAL|SPOUSE|DEPENDENTS):SSN"))
                {
                    // The manip array will not be displayed if we are breaking up a SSN into separate boxes.
                    if (subTagComb != "SSN IN 3 BOXES")
                    {
                        foreach (string str in ssnManipArray) { manipList.Items.Add(str); }
                    }
                }
                else if (Regex.IsMatch(mnemonic, @"IND\|BUSINESS:FEIN"))
                {
                    // The manip array will not be displayed if we are breaking up an FEIN into separate boxes.
                    if (subTagComb != "FEIN IN 2 BOXES")
                    {
                        foreach (string str in feinManipArray) { manipList.Items.Add(str); }
                    }
                }
                else if (Regex.IsMatch(mnemonic, @"IND\|SOLE PROPRIETOR:FEIN \/ SSN"))
                {
                    // The manip array will not be displayed if we are breaking up an FEIN or SSN into separate boxes.
                    if (subTagComb != "SEPARATE BOXES: FEIN IN 2 BOXES ELSE SSN IN 3 BOXES")
                    {
                        foreach (string str in feinSsnManipArray) { manipList.Items.Add(str); }
                    }
                }

                /*#####################################################
                #
                # Entity Profile Section:
                #
                ######################################################*/

                else if (Regex.IsMatch(mnemonic, @"ENT\|(GENERAL|AGENT):ADDRESS::ZIP"))
                {
                    foreach (string str in zipManipArray) { manipList.Items.Add(str); }
                }
                else if (Regex.IsMatch(mnemonic, @"ENT\|GENERAL:FEIN"))
                {
                    // The manip array will not be displayed if we are breaking up an FEIN into separate boxes.
                    if (subTagComb != "FEIN IN 2 BOXES")
                    {
                        foreach (string str in feinManipArray) { manipList.Items.Add(str); }
                    }
                }
                else if (Regex.IsMatch(mnemonic, @"ENT\|GENERAL\+CONTACT:PHONE"))
                {
                    foreach (string str in phoneManipArray) { manipList.Items.Add(str); }
                }

                /*#####################################################
                #
                # Equifax Profile Section:
                #
                ######################################################*/

                else if (Regex.IsMatch(mnemonic, @"EQU SUCCESSOR\|(ADDRESS|PAYROLL SERVICE PROVIDER|BANK ACCOUNT INFORMATION):(CORPORATE|MAILING|ADDRESS)::ZIP"))
                {
                    foreach (string str in zipManipArray) { manipList.Items.Add(str); }
                }
                else if (Regex.IsMatch(mnemonic, @"EQU (SUCCESSOR|PREDECESSOR)\|(OFFICERS:|CONTACT:|CORPORATE )ADDRESS::ZIP"))
                {
                    foreach (string str in zipManipArray) { manipList.Items.Add(str); }
                }
                // NOTE: We have '::' in the regular expression so that we do not erroneously match with 
                // @"EQU (SUCCESSOR|PREDECESSOR)\|OFFICERS:SSN \/ FEIN" down below, as we are not using the
                // ^ and $ characters to start and stop the match.
                else if (Regex.IsMatch(mnemonic, @"EQU (SUCCESSOR|PREDECESSOR)\|OFFICERS:SSN::"))
                {
                    // The manip array will not be displayed if we are breaking up a SSN into separate boxes.
                    if (subTagComb != "SSN IN 3 BOXES")
                    {
                        foreach (string str in ssnManipArray) { manipList.Items.Add(str); }
                    }
                }
                //Successor and Predecessor both have the same setup for FEIN list wise. Also, not including '::SINGLE BOX' at the end
                //because I am using the feinArray which only has 2 entries, "Single Box" and "FEIN in 2 boxes". This code will need changing
                //if we add more entries to feinArray.
                else if (Regex.IsMatch(mnemonic, @"EQU (SUCCESSOR|PREDECESSOR)\|(PARENT:FEIN|OFFICERS:FEIN|FEIN)"))
                {
                    // The manip array will not be displayed if we are breaking up an FEIN into separate boxes.
                    if (subTagComb != "FEIN IN 2 BOXES")
                    {
                        foreach (string str in feinManipArray) { manipList.Items.Add(str); }
                    }
                }
                else if (Regex.IsMatch(mnemonic, @"EQU SUCCESSOR\|ENTITY TYPE:DISREGARDED ENTITY::(SINGLE MEMBER FEIN|TWO BOXES: CHECK \['X'\] A BOX; SINGLE MEMBER FEIN|TWO BOXES: SINGLE MEMBER NAME; SINGLE MEMBER FEIN|THREE BOXES: CHECK \['X'\] A BOX; SINGLE MEMBER NAME; SINGLE MEMBER FEIN)"))
                {
                    // The equDisEntArray [disregarded entity] array incorporated the feinArray due to screen limitations, I could not display
                    // the feinArray in a separate list and then click on that list to see the feinManipArray.
                    foreach (string str in feinManipArray) { manipList.Items.Add(str); }
                }
                else if (Regex.IsMatch(mnemonic, @"EQU (SUCCESSOR|PREDECESSOR)\|OFFICERS:SSN \/ FEIN"))
                {
                    // The manip array will not be displayed if we are breaking up an FEIN or SSN into separate boxes.
                    if (subTagComb != "SEPARATE BOXES: FEIN IN 2 BOXES ELSE SSN IN 3 BOXES")
                    {
                        foreach (string str in feinSsnManipArray) { manipList.Items.Add(str); }
                    }
                }

                /*#####################################################
                #
                # Fiduciary Profile Section:
                #
                ######################################################*/

                else if (Regex.IsMatch(mnemonic, @"FID\|(FIDUCIARY|DECEDENT\/DONOR):ADDRESS::ZIP"))
                {
                    foreach (string str in zipManipArray) { manipList.Items.Add(str); }
                }
                else if (Regex.IsMatch(mnemonic, @"FID\|(FIDUCIARY|ESTATE\/TRUST|DECEDENT\/DONOR):SSN"))
                {
                    // The manip array will not be displayed if we are breaking up a SSN into separate boxes.
                    if (subTagComb != "SSN IN 3 BOXES")
                    {
                        foreach (string str in ssnManipArray) { manipList.Items.Add(str); }
                    }
                }
                // NOTE: We have '::' in the regular expression so that we do not erroneously match with @"FID\|ESTATE\/TRUST:FEIN \/ SSN"
                // down below, as we are not using the ^ and $ characters to start and stop the match.
                else if (Regex.IsMatch(mnemonic, @"FID\|ESTATE\/TRUST:FEIN::"))
                {
                    // The manip array will not be displayed if we are breaking up an FEIN into separate boxes.
                    if (subTagComb != "FEIN IN 2 BOXES")
                    {
                        foreach (string str in feinManipArray) { manipList.Items.Add(str); }
                    }
                }
                else if (Regex.IsMatch(mnemonic, @"FID\|ESTATE\/TRUST:FEIN \/ SSN"))
                {
                    // The manip array will not be displayed if we are breaking up an FEIN or SSN into separate boxes.
                    if (subTagComb != "SEPARATE BOXES: FEIN IN 2 BOXES ELSE SSN IN 3 BOXES")
                    {
                        foreach (string str in feinSsnManipArray) { manipList.Items.Add(str); }
                    }
                }

                /*#####################################################
                #
                # Preparer Profile Section:
                #
                ######################################################*/

                else if (Regex.IsMatch(mnemonic, @"PRE\|ADDRESS::ZIP"))
                {
                    foreach (string str in zipManipArray) { manipList.Items.Add(str); }
                }
                // NOTE: We have '::' in the regular expression so that we do not erroneously match with the other ID
                // Number mnemonic matches down below, as we are not using the ^ and $ characters to start and stop the match.
                else if (Regex.IsMatch(mnemonic, @"PRE\|ID NUMBER:FEIN::"))
                {
                    // The manip array will not be displayed if we are breaking up an FEIN into separate boxes.
                    if (subTagComb != "FEIN IN 2 BOXES")
                    {
                        foreach (string str in feinManipArray) { manipList.Items.Add(str); }
                    }
                }
                else if (Regex.IsMatch(mnemonic, @"PRE\|ID NUMBER:SSN"))
                {
                    // The manip array will not be displayed if we are breaking up a SSN into separate boxes.
                    if (subTagComb != "SSN IN 3 BOXES")
                    {
                        foreach (string str in ssnManipArray) { manipList.Items.Add(str); }
                    }
                }
                else if (Regex.IsMatch(mnemonic, @"PRE\|ID NUMBER:PTIN \/ FEIN \/ SSN"))
                {
                    // The manip array will not be displayed if we are breaking up a PTIN, a SSN, or an FEIN into separate boxes.
                    if (subTagComb != "SEPARATE BOXES: PTIN IN 1 BOX ELSE FEIN IN 2 BOXES ELSE SSN IN 3 BOXES")
                    {
                        foreach (string str in ptinFeinSsnManipArray) { manipList.Items.Add(str); }
                    }
                }
                else if (Regex.IsMatch(mnemonic, @"PRE\|ID NUMBER:PTIN \/ FEIN"))
                {
                    // The manip array will not be displayed if we are breaking up a PTIN and an FEIN into separate boxes.
                    if (subTagComb != "SEPARATE BOXES: PTIN IN 1 BOX ELSE FEIN IN 2 BOXES")
                    {
                        foreach (string str in ptinFeinManipArray) { manipList.Items.Add(str); }
                    }
                }
                else if (Regex.IsMatch(mnemonic, @"PRE\|ID NUMBER:PTIN \/ SSN"))
                {
                    // The manip array will not be displayed if we are breaking up a PTIN and a SSN into separate boxes.
                    if (subTagComb != "SEPARATE BOXES: PTIN IN 1 BOX ELSE SSN IN 3 BOXES")
                    {
                        foreach (string str in ptinSsnManipArray) { manipList.Items.Add(str); }
                    }
                }
                else if (Regex.IsMatch(mnemonic, @"PRE\|ID NUMBER:FEIN \/ PTIN"))
                {
                    // The manip array will not be displayed if we are breaking up an FEIN and a PTIN into separate boxes.
                    if (subTagComb != "SEPARATE BOXES: FEIN IN 2 BOXES ELSE PTIN IN 1 BOX")
                    {
                        foreach (string str in feinPtinManipArray) { manipList.Items.Add(str); }
                    }
                }
                else if (Regex.IsMatch(mnemonic, @"PRE\|ID NUMBER:FEIN \/ SSN"))
                {
                    // The manip array will not be displayed if we are breaking up an FEIN and a SSN into separate boxes.
                    if (subTagComb != "SEPARATE BOXES: FEIN IN 2 BOXES ELSE SSN IN 3 BOXES")
                    {
                        foreach (string str in feinSsnManipArray) { manipList.Items.Add(str); }
                    }
                }
                // NOTE: We have ':' in the regular expression so that we do not erroneously match with @"PRE\|PHONE",
                // as we are not using the ^ and $ characters to start and stop the match.
                else if (Regex.IsMatch(mnemonic, @"PRE\|PHONE:"))
                {
                    foreach (string str in phoneManipArray) { manipList.Items.Add(str); }
                }
            }
        }

        private void manipList_SelectedIndexChanged(object sender, EventArgs e)
        {
            // Return if nothing is selected in the list.
            if (manipList.SelectedIndex != -1)
            {
                if ((String)manipList.SelectedItem == "<None>")
                {
	                // Back out the non-none manip_list selection since they clicked on "<None>".
                    if (mnemonic.IndexOf(":::") > -1)
                    {
                        mnemonic = mnemonic.Substring(0, mnemonic.IndexOf(":::"));
                    }
                }
                else
                {
                    // .IndexOf(":::") returns -1 if ":::" is not found, else it returns the zero based index of the first occurrence in which case
                    // we grab all chars from the start up to but not including the ":::".
                    mnemonic = string.Format("{0}:::{1}", (mnemonic.IndexOf(":::") > -1 ? mnemonic.Substring(0, mnemonic.IndexOf(":::")) : mnemonic), ((String)manipList.SelectedItem).ToUpper());
                }
            }
        }

        #endregion

        /*
         * This inserts a mnemonic selection into the renumList. The selection is from one of two possible
         * sources: from the 'Selection List'/mnemonicListBox [a highlighted list element] or from the
         * list selections at the top of the screen.
         * 
         * Appropriate box number info will be displayed depending on the source. If the source is from
         * the 'Selection List' no mnemonic verification will be needed [it is in the associated dfObject
         * in the dfObjectList] else we will first need to verify and retrieve the mnemonic info from the
         * code snippet file.
         * 
         * The user will be required to select a single 'Selection List' element for renumbering, an error
         * message will inform them accordingly.
        */




        private void globalRenumButton_Click(object sender, EventArgs e)
        {
            string num;
            string width;
            string desc;
            string entry;
            DataFlow dfObjectCopy;

            if (mnemonicListBox.SelectedItems.Count > 0)
            {
                if (mnemonicListBox.SelectedItems.Count > 1)
                {
                    ProcessErrors("Error!", 7, null, null);
                    return;
                }
                else
                {
                    /*
                     * Here is where we rely on the 1:1 relationship between the global dfObjectList and
                     * the Selection List - the indices are the same and refer to the same info.
                     * 
                     * We are copying over dfObjectList[index num] [an object of the class DataFlow]
                     * member by member and using the new List<string>('object in dfObjectList'.[BoxWidth/
                     * NewBoxNum]) method to avoid the situation where 2 variables point to the same memory
                     * location, thus a change in 1 changes the other. We want to keep the box numbers and
                     * box widths between the renumList and the 'Selection List' [mnemonicListBox] separate
                     * memory-wise so that they can retrieve [and renumber/change widths] either the 
                     * 'Selection List' version of a mnemonic or the renum version of the mnemonic with the
                     * correct box numbers and box widths.
                     * 
                     * It boils down to copying by value NOT by reference.
                    */
                    dfObject = new DataFlow("");

                    dfObject.Mnemonic = dfObjectList[mnemonicListBox.SelectedIndex].Mnemonic;
                    dfObject.DataFlowCode = dfObjectList[mnemonicListBox.SelectedIndex].DataFlowCode;
                    dfObject.BoxWidth = new List<string>(dfObjectList[mnemonicListBox.SelectedIndex].BoxWidth);
                    dfObject.BoxNum = dfObjectList[mnemonicListBox.SelectedIndex].BoxNum;
                    dfObject.NewBoxNum = new List<string>(dfObjectList[mnemonicListBox.SelectedIndex].NewBoxNum);
                    dfObject.BoxDesc = dfObjectList[mnemonicListBox.SelectedIndex].BoxDesc;
                }
            }
            else
            {
                // No need to continue if a null mnemonic.
                if (mnemonic == null)
                {
                    return;
                }


               
                
              


                dfObjectCopy = dfObject;

                
                dfObject = GetDataflowByMnemonic(mnemonic, retMode == retrieverMode.retrieveJS);

                // Restore dfObject if: GetDataflowByMnemonic() returned null.
                if (dfObject == null)
                {
                    dfObject = dfObjectCopy;
                }
            }

            
            if (dfObject != null)
            {
                // Clear the renumber list first.
                renumList.Items.Clear();




                for (int index = 0; index < dfObject.BoxNum.Count; index++)
                {
                    /////////////////////////////////////////////////////////////////
                    // This is bit of a hack but it works. The result variable grabs
                    // the box ID which still contains the page number like box1_1.
                    // To lose the page number we first do a substring(0,3) to get
                    // just the "box" prefix. Then we use last LastIndexOf('_') which
                    // is similar to JS .split to get everything after the underscore
                    // which gives us just box number.
                    /////////////////////////////////////////////////////////////////

                    //num = dfObject.NewBoxNum[index];
                    var result = dfObject.BoxNum[index];



                    //MessageBox.Show("Showing the var result " + result);

                    if (result.Contains('_'))
                    {
                        //MessageBox.Show("Old Box Type");
                    }


                    if (checkBox1.Checked)
                    {
                      // num = ( (result.Substring(result.LastIndexOf('_') + 1)));
                       //num = (result.Substring(0, 3) + (result.Substring(result.LastIndexOf('_') + 1)));
                    }
                    else
                    {
                        //num = dfObject.BoxNum[index];
                    }


                    // 08/05/2021 - The checkbox in the selection window now defaults to a new box
                    // format if the user doesn't select a box from a boxlist. If they want an
                    // and old box number type as the default they need to uncheck the checkbox.

                    /////////////////////////////////////
                    ///Check for not boxFlag checked.
                    /////////////////////////////////////

                    // ####################
                    if (BoxFlag)
                    // ####################
                    {
                        // We have the old type boxes so box1_1 is the default.
                        //num = dfObject.BoxNum[index];
                        
                    }
                    else
                    {
                        // We have the new type of boxes so Substring out "box" and
                        // then Subsstring only the box number so box1 is the default.

                        //num = (result.Substring(0, 3) + (result.Substring(result.LastIndexOf('_') + 1)));
                       // num = dfObject.BoxNum[index];

                    }


                    //num = dfObject.BoxNum[index];
                    num = (result.Substring(0, 3) + (result.Substring(result.LastIndexOf('_') + 1)));


                    width = dfObject.BoxWidth[index];
                    desc = dfObject.BoxDesc[index];

                    /*
                    Padding the 'entry' string so it lines up with the text above the ListBox.
                    I added the tab char [\t] so we can scroll to the right for long list items and see all
                    of the visible text. I tried adding an extra space char [one or more] but it did not work.
                    I tried adding a char OTHER than a space char and it worked. Hmmm....
                    */
                    entry = (string.Format("{0}{1}{2}{3}{4}\t", num, new string(' ', (num.Length > 10 ? 0 : 10 - num.Length)), width, new string(' ', (width.Length > 11 ? 0 : 11 - width.Length)), desc));
                    renumList.Items.Add(entry);
                }


                
                // Select (show highlighting) first list element:
                renumList.SetSelected(0, true);

                // Activate the 'Truncate' entry field:
                renumTruncPanel.Enabled = true;
            }
        }

        /*
        This handles the renumbering of the boxes associated with a mnemonic selection. Multiple selections can be renumbered
        in one click as long as the same number of boxes are selected in the formBoxListBox to aid in renumbering speed.
        */
        private void renumModifyButton_Click(object sender, EventArgs e)
        {
            string num;
            string width;
            string desc;
            string entry;
            List<int> renumListSelectedIndicesCopy = new List<int>();

            // Return if nothing is selected in the list.
            if (renumList.SelectedIndex != -1)
            {
                if (formBoxListBox.SelectedIndices.Count == renumList.SelectedIndices.Count)
                {
                    /*
		            The renumList has the same number of elements as the lists in the global object dfObject. The index values
		            between the two are thus 1:1.
		            */
                    for (int index = 0; index < formBoxListBox.SelectedIndices.Count; index++)
                    {
                        dfObject.NewBoxNum[renumList.SelectedIndices[index]] = formBoxListBox.Items[formBoxListBox.SelectedIndices[index]].ToString();

                        // Making a copy of the renumList selected indices, we will re-select these items when re-displaying the list.
                        renumListSelectedIndicesCopy.Add(renumList.SelectedIndices[index]);
                    }

                    //  Now clear the renumList and re-display it with the new box numbers:
                    renumList.Items.Clear();

                    for (int index = 0; index < dfObject.NewBoxNum.Count; index++)
                    {


                        var result = dfObject.NewBoxNum[index];

                        
                        if (checkBox1.Checked)
                        {
                           
                            num = (result.Substring(0, 3) + (result.Substring(result.LastIndexOf('_') + 1)));
                        }
                        else
                        {
                            num = dfObject.NewBoxNum[index];
                        }



                        width = dfObject.BoxWidth[index];
                        desc = dfObject.BoxDesc[index];

                        /*
                        Padding the 'entry' string so it lines up with the text above the ListBox.
                        I added the tab char [\t] so we can scroll to the right for long list items and see all
                        of the visible text. I tried adding an extra space char [one or more] but it did not work.
                        I tried adding a char OTHER than a space char and it worked. Hmmm....
                        */
                        entry = (string.Format("{0}{1}{2}{3}{4}\t", num, new string(' ', (num.Length > 10 ? 0 : 10 - num.Length)), width, new string(' ', (width.Length > 11 ? 0 : 11 - width.Length)), desc));
                        renumList.Items.Add(entry);
                    }

                    // Re-highlight the same renumList elements:
                    for (int index = 0; index < renumListSelectedIndicesCopy.Count; index++)
                    {
                        renumList.SetSelected(renumListSelectedIndicesCopy[index], true);
                    }
                }
                else
                {
                    ProcessErrors("Error!", 6, null, null);
                }
            }

        }


        
        /*
        This clears [deletes] the renumList and also resets the global dfObject to null.
        
        We will make sure there is something in the renumList first so that we do not set dfObject to null incorrectly.
        What could happen is that dfObject holds mnemonic info but the renumList was not used in obtaining that info
        [in which case the list would be empty], perhaps this may not happen, this is just a safeguard.
        */
        private void renumClearButton_Click(object sender, EventArgs e)
        {
            if (renumList.Items.Count > 0)
            {
                renumList.Items.Clear();
                dfObject = null;

                // De-activate the Manual Truncate entry field and clear it out:
                manTruncEntry.Clear();
                renumTruncPanel.Enabled = false;

		        // Put the cursor back in the retrieveText RichTextBox because the Manual Truncate
		        // entry widget may have had it if they used it.
                retrieveText.Focus();
            }
        }

        /*
        This adds truncation width to the selected box(es) displayed in the renumList for APF form dataflow truncation.
        The width either comes from the manTruncEntry [manual truncation] field if a positive value is there or it
        will come from the formBoxListBox selection which may possibly be set at the XML file read operation to "MAN"
        indicating that the font is not the correct type for automatic calculation of box width [in chars] and the user
        will need to manually enter the truncation length value.
        */
        private void truncAddButton_Click(object sender, EventArgs e)
        {
            int manTruncValue = 0;
            string num;
            string width;
            string desc;
            string entry;
            List<int> renumListSelectedIndicesCopy = new List<int>();

            // Return if nothing is selected in the list.
            if (renumList.SelectedIndex != -1)
            {
                for (int index = 0; index < renumList.SelectedIndices.Count; index++)
                {
                    // Making a copy of the renumList selected indices, we will re-select these items when re-displaying the list.
                    renumListSelectedIndicesCopy.Add(renumList.SelectedIndices[index]);
                }

                if (manTruncEntry.Text.Length > 0)
                {
                    if (Regex.IsMatch(manTruncEntry.Text, @"^\d{1,3}$"))
                    {
                        int.TryParse(manTruncEntry.Text, out manTruncValue);
                    }
                    else
                    {
                        ProcessErrors("Error!", 13, manTruncEntry.Text, "Truncate");

                        // Set the cursor in the manual truncation entry text box since they used it.
                        manTruncEntry.Select();

                        return;
                    }
                }

                if (manTruncValue > 0)
                {
                    // The truncation value is manually entered, so we will use it:
                    for (int index = 0; index < renumList.SelectedIndices.Count; index++)
                    {
                        dfObject.BoxWidth[renumList.SelectedIndices[index]] = manTruncEntry.Text;
                    }

                    manTruncEntry.Text = "";

                    // Set the cursor in the manual truncation entry text box since they used it.
                    manTruncEntry.Select();
                }
                else
                {
		            // The truncation value(s) will be obtained from the formBoxList by way of the formBoxListBox.
                    if (formBoxListBox.SelectedIndices.Count == renumList.SelectedIndices.Count)
                    {
                        for (int index = 0; index < formBoxListBox.SelectedIndices.Count; index++)
                        {
                            // Start Debug:
                            //if (dfObject == null)
                            //{
                            //    MessageBox.Show("ADD: dfObject is null!!", "Debug:", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                            //}
                            // End Debug.

                            dfObject.BoxWidth[renumList.SelectedIndices[index]] = formBoxList[formBoxListBox.SelectedIndices[index]].Width;
                        }
                    }
                    else
                    {
                        ProcessErrors("Error!", 6, null, null);
                        return;
                    }
                }

	            // Things went well up above [else we would not be here], so now clear the renumList and re-display it with
	            // the new box width(s):
                renumList.Items.Clear();

                for (int index = 0; index < dfObject.NewBoxNum.Count; index++)
                    {
                    
                    var result = dfObject.NewBoxNum[index];
                    
                    if (checkBox1.Checked)
                    {

                        //num = (result.Substring(0, 3) + (result.Substring(result.LastIndexOf('_') + 1)));
                    }
                    else
                    {
                        //num = dfObject.NewBoxNum[index];
                    }


                    width = dfObject.BoxWidth[index];
                    desc = dfObject.BoxDesc[index];


                    /*
                    Padding the 'entry' string so it lines up with the text above the ListBox.
                    I added the tab char [\t] so we can scroll to the right for long list items and see all
                    of the visible text. I tried adding an extra space char [one or more] but it did not work.
                    I tried adding a char OTHER than a space char and it worked. Hmmm....
                    */

                    num = (num.Substring(num.LastIndexOf('_') + 1));

                    entry = (string.Format("{0}{1}{2}{3}{4}\t", num, new string(' ', (num.Length > 10 ? 0 : 10 - num.Length)), width, new string(' ', (width.Length > 11 ? 0 : 11 - width.Length)), desc));
                    renumList.Items.Add(entry);
                }

                // Re-highlight the same renumList elements:
                for (int index = 0; index < renumListSelectedIndicesCopy.Count; index++)
                {
                    renumList.SetSelected(renumListSelectedIndicesCopy[index], true);
                }
            }
        }

        /*
        This deletes truncation widths from the selected box(es) displayed in the renum_list.
        */
        private void truncDelButton_Click(object sender, EventArgs e)
        {
            string num;
            string width;
            string desc;
            string entry;
            List<int> renumListSelectedIndicesCopy = new List<int>();

            // Return if nothing is selected in the list.
            if (renumList.SelectedIndex != -1)
            {
                for (int index = 0; index < renumList.SelectedIndices.Count; index++)
                {
                    // Start Debug:
                    //if (dfObject == null)
                    //{
                    //    MessageBox.Show("DEL: dfObject is null!!", "Debug:", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                    //}
                    // End Debug.

                    dfObject.BoxWidth[renumList.SelectedIndices[index]] = "";

                    // Making a copy of the renumList selected indices, we will re-select these items when re-displaying the list.
                    renumListSelectedIndicesCopy.Add(renumList.SelectedIndices[index]);
                }

	            // Now clear the renum_list and re-display it with the selected width(s) deleted:
                renumList.Items.Clear();

                for (int index = 0; index < dfObject.NewBoxNum.Count; index++)
                {
                    
                    var result = dfObject.NewBoxNum[index];


                    if (checkBox1.Checked)
                    {

                        num = (result.Substring(0, 3) + (result.Substring(result.LastIndexOf('_') + 1)));
                    }
                    else
                    {
                        num = dfObject.NewBoxNum[index];
                    }




             /*      

                    // ####################
                    if (BoxFlag)
                    // ####################
                    {
                        // We have the old type boxes so box1_1 is the default.
                        num = dfObject.NewBoxNum[index];
                    }
                    else
                    {
                        // We have the new type of boxes so Substring out "box" and
                        // then Subsstring only the box number so box1 is the default.
                        
                        num = (result.Substring(0, 3) + (result.Substring(result.LastIndexOf('_') + 1)));
                        //num = (result);
                    }

            */


                    width = dfObject.BoxWidth[index];
                    desc = dfObject.BoxDesc[index];

                    /*
                    Padding the 'entry' string so it lines up with the text above the ListBox.
                    I added the tab char [\t] so we can scroll to the right for long list items and see all
                    of the visible text. I tried adding an extra space char [one or more] but it did not work.
                    I tried adding a char OTHER than a space char and it worked. Hmmm....
                    */
                    entry = (string.Format("{0}{1}{2}{3}{4}\t", num, new string(' ', (num.Length > 10 ? 0 : 10 - num.Length)), width, new string(' ', (width.Length > 11 ? 0 : 11 - width.Length)), desc));
                    renumList.Items.Add(entry);
                }

                // Re-highlight the same renumList elements:
                for (int index = 0; index < renumListSelectedIndicesCopy.Count; index++)
                {
                    renumList.SetSelected(renumListSelectedIndicesCopy[index], true);
               }
            }
        }

        /*
        This inserts the current mnemonic variable contents into the Mnemonic Composition
        List [the GUI refers to this as 'Selection List', internally this is a ListBox named
        mnemonicListBox].
        
        There are two possible ways the current mnemonic selection will be inserted:
        
        1.) If the current mnemonic selection is also present in the 'Renumber and/or Truncate
        Selection Boxes' list, then the mnemonic will be inserted using the box numbers from that
        list. Also, since we would be inserting from the list, we already know that the mnemonic has been
        built-up properly [i.e. it was located in the snippet file and has been retrieved already,
        dataflow code too].
        
        2.) If the mnemonic selection is not in the 'Renumber and/or Truncate Selection Boxes' list,
        we will first verify that it has been properly built, and it will then be displayed in the
        'Selection List' showing the default box numbers from the snippet file. If the verification fails,
        an error message will inform the user and the operation will be aborted.
        
        The 'Selection List' is implemented as a list of (dfObject)s, so no further interaction with
        the code snippet file will be required.
        
        If a list element is selected [or several], the insertion will occur either above, over,
        or below the selection(s), based on the current toggle state of the Insert Above/Over/Below
        button.
        
        If a list element is not selected, the insertion will either be at the beginning for Insert
        Above, at the end of the list for Insert Below, or the list will be deleted and then the mnemonic
        will be added to the list for Insert Over.
        
        An error message will be displayed if non-contiguous list elements are selected.
        */
        private void insertMnemonicButton_Click(object sender, EventArgs e)
        {
            bool resetSelections = false;
            int flag = 0;
            int errNum = 0;
            int widthInt;
            int insertIndex = 0;
            string selection;
            string box;
            string width;
            List<string> tmpList = new List<string>();

            /*
	        First determine if we are inserting the mnemonic from the 'Renumber and/or Truncate
            Selection Boxes' list [renumList] or if we are using the currently selected mnemonic per
            the list selections on the top area of the screen.
	        
	        If the mnemonic is in the renumList, no verification of the mnemonic will be required
	        as that has already been done for insertion into that list.
	        
	        If we are using the mnemonic as built from the list selections [top area of screen]
	        GetDataflowByMnemonic( ) will inform user via an error message if the mnemonic has not been
	        properly formed, in which case this routine will abruptly end.
            */
            if (dfObject == null)
            {
                // But no sense in continuing if a null mnemonic....
                if (mnemonic == null)
                {
                    return;
                }
                

                dfObject = GetDataflowByMnemonic(mnemonic, retMode == retrieverMode.retrieveJS);

                if (dfObject == null)
                {
                    return;
                }
            }

            if (mnemonicListBox.SelectedIndices.Count > 0)
            {
		        // One or more list elements are selected. We need to check for a
		        // contiguous selection before we can continue.
                for (int index = mnemonicListBox.SelectedIndices[0]; index <= mnemonicListBox.SelectedIndices[mnemonicListBox.SelectedIndices.Count - 1]; index++)
                {
                    if (!mnemonicListBox.SelectedIndices.Contains(index))
                    {
                        // Selection is non-contiguous. Popup an error message and exit the routine.
				        ProcessErrors("Error!", 3, null, null);
                        return;
                    }
                }

                // 20 indicates a contiguous selection which is good.
                flag = 20;

               
            

                // We will reset the selected items in the list at the end of this routine
                // [not doing it before that because we may exit early due to an error].
                resetSelections = true;
            }

	        /*
	        Position the horizontal scrollbar to the default [far left] position so they can clearly see
	        the elements. The scrollbar could be way to the right because they were viewing a long
	        mnemonic string and we want them to be able to see what they are inserting.
            
            In converting to C#, this is the only way I could get this to work.
	        */
            mnemonicListBox.RightToLeft = RightToLeft.Yes;
            mnemonicListBox.RightToLeft = RightToLeft.No;

            /*
	        Now check if the mnemonic is already in the list - this may or may not be a problem - 
	        we need to check the Insert button's state further down below. Each 'selection'
	        has a box list prefix [each box may also have a truncation value if applicable] followed
	        by the mnemonic, so we need to split apart each list element inserting the mnemonic part
	        into tmpList then check if the mnemonic we are inserting is in tmpList and process
	        accordingly.
	        */
            for (int index = 0; index < mnemonicListBox.Items.Count; index++)
            {
                // We split each item based on " || ", which gives us a two element array per split. The 
                // second array element is the mnemonic. We control the strings in the mnemonicListBox,
                // so we always know there is a " || " component per list item.


               



                tmpList.Add(mnemonicListBox.Items[index].ToString().Split(new string[] { " || " }, StringSplitOptions.None)[1]);


            }

            if (tmpList.Contains(dfObject.Mnemonic))
            {
                flag += 1;
            }
            


	        /*
	        We now add a box list prefix string to the mnemonic. We use the NewBoxNum list
	        box values in dfObject for the prefix, boxes are separated by a " " [space] character.
	        If we are truncating a box, we will append the length in brackets to the box number,
	        so truncating box1_13 at 24 characters would yield "box1_13[24]".
	        
	        We will insert selection into the list if appropriate, we will use mnemonic for 
	        ProcessErrors( ) as no need in having the box num prefix data in the error window.
	        */
            selection = "";
            

            for (int index = 0; index < dfObject.NewBoxNum.Count; index++)
            {
             
                var result = dfObject.NewBoxNum[index];
                
                   if (checkBox1.Checked)
                   {

                       box = ((result.Substring(result.LastIndexOf('_') + 1)));
                   }
                   else
                   {
                       box = dfObject.NewBoxNum[index];
                   }

                   

                //box = (result.Substring(0, 3) + (result.Substring(result.LastIndexOf('_') + 1)));
                //box = dfObject.NewBoxNum[index];

                // We have complete control over the width's: They are either "MAN", "NO XML" or integer
                // values expressed as strings. TryParse will return 0 for "MAN" and "NO XML" values.
                int.TryParse(dfObject.BoxWidth[index], out widthInt);

                width = (widthInt > 0 ? String.Format("[{0}]", widthInt) : "");
                selection += String.Format("{0}{1} ", box, width);

               selection = String.Format("{0}|| {1}", selection, dfObject.Mnemonic);

                
            }
            
            //selection = (selection.Substring(0, 3) + (selection.Substring(selection.LastIndexOf('_') + 1)));

            //selection = (selection.Substring(selection.LastIndexOf('_') + 1));

            

            /*
	        At this point, flag is either 0, 1, 20, or 21. We need to now take the Insert button
	        state into account before we can insert the selection into the list.
	        Also, references to 'selection' in the below case expression are just the mnemonic
	        with the box list prefix.
            
            flag values: 0 => The mnemonic is not in the list, there are no selections in the list;
            1 => The mnemonic is already in the list; 20 => There is one or more contiguous [no gaps] selections;
            21 => There is a contiguous selection(s) and the mnemonic is in the list already.
	        */
            if (insertAboveMnemonicButton.BackColor == SystemColors.Info)
            {
                /*
                Inserting Above:
                
                Inserting "Above": Inserts mnemonic with box prefix info into the 'Selection List' "Above" the
                insertion point which is either a selected list element, or a contiguous range of selected
                elements, or at the top of the list.
                */

               

                switch (flag)
                {
                    case (0):
                        mnemonicListBox.Items.Insert(0, selection);
                        break;

                    case (20):
                        mnemonicListBox.Items.Insert(mnemonicListBox.SelectedIndices[0], selection);
                        break;

                    default:
                        if (Regex.IsMatch(flag.ToString(), @"^2?1$"))
                        {
                            errNum = 2;
                        }
                        break;
                }
            }
            else if (insertOverMnemonicButton.BackColor == SystemColors.Info)
            {
                /*
                Inserting Over:
                
                Inserting "Over": Inserts mnemonic with box prefix info into the 'Selection List' "Over" the
                insertion point which is either a selected list element, or a contiguous range of selected
                elements, or the list is first emptied and then the mnemonic will be inserted and will be the only
                item in the list.
                */
                switch (flag)
                {
                    case (20):

					    /*
					    List items are selected (contiguously), there is no duplicate mnemonic
					    in the list already, so we insert the selection over the selected item(s)
					    thus they will be replaced by the single selection.
					    */
                        insertIndex = mnemonicListBox.SelectedIndices[0];

                        for (int index = mnemonicListBox.SelectedIndices.Count - 1; index >= 0; index--)
                        {
                            mnemonicListBox.Items.RemoveAt(mnemonicListBox.SelectedIndices[index]);
                        }

                        mnemonicListBox.Items.Insert(insertIndex, selection);
                        mnemonicListBox.SetSelected(insertIndex, true);
                        break;

                    case (21):

					    /*
					    List items are selected (contiguously), there is a duplicate mnemonic
					    in the list already, so we can only insert IF the duplicate mnemonic
					    resides inside the selected area we are to overwrite. We popup an
					    error message if the duplicate mnemonic resides outside the selected
					    area.
					    */
                        if (Enumerable.Range(mnemonicListBox.SelectedIndices[0], mnemonicListBox.SelectedIndices.Count).Contains(tmpList.IndexOf(dfObject.Mnemonic)))
                        {
                            insertIndex = mnemonicListBox.SelectedIndices[0];

                            for (int index = mnemonicListBox.SelectedIndices.Count - 1; index >= 0; index--)
                            {
                                mnemonicListBox.Items.RemoveAt(mnemonicListBox.SelectedIndices[index]);
                            }

                            mnemonicListBox.Items.Insert(insertIndex, selection);
                            mnemonicListBox.SetSelected(insertIndex, true);
                        }
                        else
                        {
                            errNum = 2;
                        }
                        break;

                    default:
                        if (Regex.IsMatch(flag.ToString(), @"^[01]$"))
                        {
                            /*
					        No selections in list so we delete the list and simply insert the
					        selection at the 'top' of the list - we do not care if the mnemonic
					        was already in the list:
					        */
                            

                            mnemonicListBox.Items.Clear();
                            mnemonicListBox.Items.Add(selection);
                        }
                        break;
                }
            }
            else if (insertBelowMnemonicButton.BackColor == SystemColors.Info)
            {
                /*
                Inserting Below:
                
                Inserting "Below": Inserts mnemonic with box prefix info into the 'Selection List' "Below" the
                insertion point which is either a selected list element, or a contiguous range of selected
                elements, or at the bottom of the list.
                */
                switch (flag)
                {
                    case (0):
                        mnemonicListBox.Items.Add(selection);
                        break;

                    case (20):
                        mnemonicListBox.Items.Insert(mnemonicListBox.SelectedIndices[mnemonicListBox.SelectedIndices.Count - 1] + 1, selection);
                        break;

                    default:
                        if (Regex.IsMatch(flag.ToString(), @"^2?1$"))
                        {
                            errNum = 2;
                        }
                        break;
                }
            }

            if (errNum > 0)
            {
                ProcessErrors("Error!", errNum, dfObject.Mnemonic, null);

                // Do not want to kill this if it is linked to the renum list.
                if (renumList.Items.Count == 0)
                {
                    dfObject = null;
                }
            }
            else
            {
                /*
                We inserted the selection, now we need to do the following:
		        
                1) Re-synch the global dfObjectList with the 'Selection List' to keep them 1:1. [SynchDfObjList( )]
                2) Clear the 'Renumber and/or Truncate Selection Boxes' list so user does not get confused with it still being up there.
                3) Clear dfObject [set it to null].
                4) Make sure the new insertion is visible in the window.
                */
                dfObjectList = SynchDfObjList(dfObject, mnemonicListBox, dfObjectList, true);

                dfObject = null;

		        /*
		        If the renumList is not empty we will empty the list, deactivate the manual truncation field,
		        and clear the manual truncation field.
		        */
                if (renumList.Items.Count > 0)
                {
                    renumList.Items.Clear();
                    manTruncEntry.Clear();
                    renumTruncPanel.Enabled = false;

                    // Setting focus to the dataflow retrieval area.
                    retrieveText.Select();
                }

		        // We are currently resetting the selections in the mnemonicListBox after a successful insertion.
                if (resetSelections)
                {
                    mnemonicListBox.ClearSelected();
                }

		        // Making sure they can 'see' the new list insertion [in case we scroll out of view]:
                mnemonicListBox.TopIndex = mnemonicListBox.Items.Count - 1;
            }
        }

        /*
        Insert "Above" mnemonic control button:
        
        Only one Insertion control button will be active at a time, the default will be the Insert Below
        control button.
        
        Inserting "Above": Inserts mnemonic with box prefix info into the 'Selection List' "Above" the
        insertion point which is either a selected list element, or a contiguous range of selected
        elements, or at the top of the list.
        */
        private void insertAboveMnemonicButton_Click(object sender, EventArgs e)
        {
          	if (insertAboveMnemonicButton.BackColor != SystemColors.Info)
            {
		        insertAboveMnemonicButton.BackColor = SystemColors.Info;
		        insertOverMnemonicButton.BackColor = Color.LightGray;
                insertBelowMnemonicButton.BackColor = Color.LightGray;
            }
        }

        /*
        Insert "Over" mnemonic control button:
        
        Only one Insertion control button will be active at a time, the default will be the Insert Below
        control button.
        
        Inserting "Over": Inserts mnemonic with box prefix info into the 'Selection List' "Over" the
        insertion point which is either a selected list element, or a contiguous range of selected
        elements, or the list is first emptied and then the mnemonic will be inserted and will be the only
        item in the list.
        */
        private void insertOverMnemonicButton_Click(object sender, EventArgs e)
        {
            if (insertOverMnemonicButton.BackColor != SystemColors.Info)
            {
                insertOverMnemonicButton.BackColor = SystemColors.Info;
                insertAboveMnemonicButton.BackColor = Color.LightGray;
                insertBelowMnemonicButton.BackColor = Color.LightGray;
            }
        }

        /*
        Insert "Below" mnemonic control button:
        
        Only one Insertion control button will be active at a time, the default will be the Insert Below
        control button.
        
        Inserting "Below": Inserts mnemonic with box prefix info into the 'Selection List' "Below" the
        insertion point which is either a selected list element, or a contiguous range of selected
        elements, or at the bottom of the list.
        */
        private void insertBelowMnemonicButton_Click(object sender, EventArgs e)
        {
            if (insertBelowMnemonicButton.BackColor != SystemColors.Info)
            {
                insertBelowMnemonicButton.BackColor = SystemColors.Info;
                insertAboveMnemonicButton.BackColor = Color.LightGray;
                insertOverMnemonicButton.BackColor = Color.LightGray;
            }
        }

        /*
        This selects all mnemonics in the mnemonicListBox.
        */
        private void selectAllMnemonicButton_Click(object sender, EventArgs e)
        {
            for (int index = 0; index < mnemonicListBox.Items.Count; index++)
            {
                mnemonicListBox.SetSelected(index, true);
            }
        }

        /*
        This will delete mnemonic elements from the 'Selection List', aka mnemonicListBox.
        If one or more elements are selected, they will be deleted, else the entire list will be emptied.
        */
        private void deleteMnemonicButton_Click(object sender, EventArgs e)
        {
            List<string> tmpList = new List<string>();

            if ((mnemonicListBox.SelectedIndices.Count == 0) || (mnemonicListBox.SelectedIndices.Count == mnemonicListBox.Items.Count))
            {
                // # Deleting the whole list:
                mnemonicListBox.Items.Clear();
            }
            else
            {
                for (int index = mnemonicListBox.SelectedIndices.Count - 1; index >= 0; index--)
                {
                    mnemonicListBox.Items.RemoveAt(mnemonicListBox.SelectedIndices[index]);
                }
            }

	        /*
	        Now we need to re-synchronize the global dfObjectList with the current state of the 'Selection List' since we
	        just did some deletions.
	        */
            dfObjectList = SynchDfObjList(dfObject, mnemonicListBox, dfObjectList, false);
        }

        /*
        This resets [clears] the current selection(s) in the mnemonicListBox ['Selection List'].
        */
        private void resetMnemonicButton_Click(object sender, EventArgs e)
        {
            mnemonicListBox.ClearSelected();
        }

        /*
         * The Retriever now operates under 3 modes:
         * 
         * Retrieve Ruby code - currently the default mode.
         * Retrieve JavaScript code
         * Convert Ruby Code to JavaScript code
         * 
         * Retrieve Ruby code and Retrieve JavaScript code both use the mnemonic string built
         * from user list selections like it has been doing all along.
         * 
         * Convert Ruby Code to JavaScript code processes Ruby code that has first been copied
         * to the clipboard.
         * 
         * So we check what mode the Retriever is in and then process accordingly.
        */
        private void retrieveButton_Click(object sender, EventArgs e)
        {
            
            switch (retMode)
            {
               

                case (retrieverMode.retrieveJS):

                    RetrieveDataflow();
                    break;

               // case (retrieverMode.convertRubyToJS):

                //    ProcessRubyConversion();
                //    break;
            }
            
            switch (retMode)
            {


                case (retrieverMode.retrieveRuby):

                    RetrieveDataflow();
                    break;

               // case (retrieverMode.convertRubyToJS):

                //    ProcessRubyConversion();
                //    break;
            }


           


        }

        /*
         * What happens here depends on what mode the application is in.
         * 
         * Retrieve Ruby Code Mode:
         * 
         *      This inserts the two global assignments [$,="\x02" and $\="\x03\n"] into retrieveText
         *      that every form with Ruby calcs must have. The insertion is made at the beginning of the
         *      text. We do not update the insertion cursor, that is, we 'put it back' where we found it
         *      [offset by the number of chars we just inserted]. We preserve any selection of text.
         *      
         * Retrieve JS Mode:
         * Convert Ruby To JS Mode:
         * 
         *      The button text will be toggled between 'OFF/ON' or 'ON/OFF', depending on which of these 
         *      modes is active.
        */
        private void generateGlobalsButton_Click(object sender, EventArgs e)
        {
           
            if (retMode == retrieverMode.retrieveRuby)
            {
                int selStart;
                int selLength;
                int globalsLength;
                string globals = "$,=\"\\x02\"\n$\\=\"\\x03\\n\"\n\n";

                string globalsReg = @"\s*\$,=""\\x02""\s*\$\\=""\\x03\\n""\s*";
                Match match = Regex.Match(retrieveText.Text, globalsReg);
            
                if (match.Success)
                {
                    /*
                     * Had to 'direct-code' globals here because 2 extra space chars were
                     * required between the lines to make them appear lined up looking at
                     * the left side.
                     */
                    ProcessErrors("Error!", 18, "$,=\"\\x02\"\n  $\\=\"\\x03\\n\"\n\n", null);
                    retrieveText.Focus();
                    return;
                }

                retrieveText.Focus();
                selStart = retrieveText.SelectionStart;
                selLength = retrieveText.SelectionLength;
                globalsLength = globals.Length;
            
                /*
                 * Insert at the start of retrieveText. Doing it in this order seems to keep the Undo
                 * and Redo functionality intact.
                */
                retrieveText.SelectionStart = 0;
                retrieveText.SelectionLength = 0;
                retrieveText.SelectedText = globals;

                // Restore cursor and any previously selected text:
                retrieveText.SelectionStart = selStart + globalsLength;
                retrieveText.SelectionLength = selLength;
            }
            else if (generateGlobalsButton.Text == "Gen Variables: OFF")
            {
                generateGlobalsButton.Text = "Gen Variables: ON";
            }
            else
            {
                generateGlobalsButton.Text = "Gen Variables: OFF";
            }
    
        }
    

        /*
         * This button is disabled when Converting from Ruby to JavaScript, so we only check if we are
         * retrieving Ruby code or retrieving Javascript code.
         * 
         * For JavaScript Retrieval:
         * 
         *  The profile declaration will be inserted at the start of the text. We do not control
         *  the positioning after that, so user beware.
         * 
         * For Ruby Retrieval:
         * 
         *  The profile declaration will be handled as it always has been, inserted at the current
         *  insertion point.
         *  
         * For either Retrieval: We preserve any selection of text.
        */
        private void generateProfileButton_Click(object sender, EventArgs e)
        {
            int selStart;
            int selLength;
            string profileStr;

	        if (String.IsNullOrEmpty(profile))
            {
		        ProcessErrors("Error!", 0, null, null);
            }
	        else
            {
                if (retMode == retrieverMode.retrieveJS)
                {
                    profileStr = GetJSSpecialCode(profile.ToUpper());
                }
                else
                {
                    profileStr = string.Format("{0} = {1}.new\n", profile.ToLower(), profile);
                }

                if (!String.IsNullOrEmpty(profileStr))
                {
                    // First check for a duplicate profile declaration:
                    string retStr = Regex.Replace(retrieveText.Text, @"\s+", "");
                    string profStr = Regex.Replace(profileStr, @"\s+", "");

                    if (retStr.Contains(profStr))
                    {
                        ProcessErrors("Error!", 17, profile, profileStr);
                        retrieveText.Focus();
                        return;
                    }

                    selStart = retrieveText.SelectionStart;
                    selLength = retrieveText.SelectionLength;

                    if (retMode == retrieverMode.retrieveJS)
                    {
                        /*
                         * Insert at the start of retrieveText. Doing it in this order seems to keep the
                         * Undo and Redo functionality intact. We append a newline if the text box starts
                         * with text that is not another profile instantiation or there is no text
                         * currently in the text box.
                        */
                        retrieveText.SelectionStart = 0;
                        retrieveText.SelectionLength = 0;

                        if (Regex.IsMatch(retrieveText.Text, @"^\s*\w"))
                        {
                            // Not sure if var or let will be used, so looking for either.
                            if (!Regex.IsMatch(retrieveText.Text, @"^\s*(var|let)\s+(entity|individual|preparer|fiduciary|equifax)\s+=\s+"))
                            {
                                profileStr += "\n";
                            }
                        }
                        else if (retrieveText.Text.Length == 0)
                        {
                            profileStr += "\n";
                        }

                        retrieveText.SelectedText = profileStr;

                        // Restore cursor and any previously selected text:
                        retrieveText.SelectionStart = selStart + profileStr.Length - 1;
                        retrieveText.SelectionLength = selLength;
                    }
                    else if (retMode == retrieverMode.retrieveRuby)
                    {
                         //If there is any selected text, this will insert after that:
                        retrieveText.SelectionStart += retrieveText.SelectionLength;

                        // Temporarily 'turn off' any selected text, if any:
                        retrieveText.SelectionLength = 0;

                        retrieveText.SelectedText = profileStr;

                        /*
                         * If text was selected [selLength > 0]: Restore any selection of text,
                         * we are stuck leaving the cursor where it was originally, with selected
                         * text it is at the end and is not visible.
                         */
                        retrieveText.SelectionStart = (selLength > 0 ? selStart : selStart + profileStr.Length);
                        retrieveText.SelectionLength = selLength;
                    }
                }
            }

            retrieveText.Focus();
        }

        /*
        This deletes selected text or all text if no text is selected.
        */
        private void deleteDataflowButton_Click(object sender, EventArgs e)
        {
            if (retrieveText.SelectedText.Length > 0)
            {
                retrieveText.SelectedText = "";
            }
            else
            {
                retrieveText.Clear();
            }

            retrieveText.Focus();
        }

        /*
         * This inserts conditional code into the retrieveText area with the selected text inside the 
         * conditional body via tabbing. For example using Ruby, an 'if (something is true)'
         * conditional would encapsulate the selected text between the 'if (..)' and the 'end'.
         * 
         * NOTE: THIS BUTTON IS DISABLED WHEN CONVERTING RUBY TO JS.
        */
        private void condButton_Click(object sender, EventArgs e)
        {
            string selText = "";
            string retrieveStr = "";
            string[] linesArray;
            CondExp condExp = null;
            Regex regex;

            if (condList.Items.Count > 0)
            {
                if (condList.SelectedIndices.Count > 0)
                {
                    // Good so far, a conditional operation has been selected.

                    selText = retrieveText.SelectedText;

                    if (selText.Length > 0)
                    {
                        // We now make sure they made a "good" text selection:
                        if (!GoodSelection(ref selText))
                        {
                            return;
                        }

                        // Since we made it here we know that they made a good selection.
                        condExp = GetCond(profileTag, condList);

                        // Adding some defensive coding here, THIS SHOULD NEVER HAPPEN, but.....
                        if (condExp == null)
                        {
                            ProcessErrors("Error! Tell Harry!!", 14, string.Format("GetCond(\"{0}\", condList)  \n\n  Which was called from:\n\n  condButton_Click()", profileTag), null);
                            retrieveText.Focus();
                            return;
                        }

                        // We start building up retrieveStr, making it one long string.
                        retrieveStr = condExp.IfStr;

                        /*
                         * Before we .Split on newlines, we change the instances of newlines to a space
                         * char followed by a newline, thus preserving blank lines in the selected text
                         * [since .Split below would not permit empty entries].
                        */
                        regex = new Regex(@"\n");
                        selText = regex.Replace(selText, " \n");

                        linesArray = selText.Split(new string[] { "\r\n", "\n" }, StringSplitOptions.RemoveEmptyEntries);

                        foreach (string line in linesArray)
                        {
                            retrieveStr += string.Format("\t{0}\n", line);
                        }

                        /*
                         * The 'else' part of the conditional expression if aplicable:
                        */
                        if (condExp.ElseStr != null)
                        {
                            retrieveStr += condExp.ElseStr;
                        }

                        // The 'end' of the conditional expression - depends on if Ruby or JS:
                        retrieveStr += string.Format("{0}\n", (retMode == retrieverMode.retrieveJS ? "}" : "end"));

                        /*
                         * Now we are ready to 'insert' the selected code that we grabbed and tabbed and
                         * processed for conditional processing back into retrieveText, replacing the
                         * selected code. First we need to back out those space chars we inserted in front
                         * of the newline chars up above.
                        */
                        regex = new Regex(@" \n");
                        retrieveStr = regex.Replace(retrieveStr, "\n");
                        retrieveText.SelectedText = retrieveStr;

                        // Now clear the conditional selection:
                        condList.ClearSelected();
                    }
                    else
                    {
                        ProcessErrors("Error!", 12, null, null);
                    }
                }
                else
                {
                    ProcessErrors("Error!", 15, null, null);
                }
            }

            retrieveText.Focus();
        }

        /*
        This renumbers a selected box in the retrieveText area using a selected box from the Box List.
        Only one box may be selected in either area at a time. No ranges.
        */
        private void renumBoxButton_Click(object sender, EventArgs e)
        {
            Regex regex;

            if (retrieveText.SelectedText.Length > 0)
            {
                // They selected text, now we inspect it and make sure the text contains 1 box and
                // they selected 1 box from 'Box List' [formBoxListBox].
                if ((Regex.Matches(retrieveText.SelectedText, @"box\d+_\d+|box\d+").Count != 1) || (formBoxListBox.SelectedIndices.Count != 1))
                {
                    ProcessErrors("Error!", 10, null, null);
                }
                else
                {
                    regex = new Regex(@"box\d+_\d+|box\d+");
                    retrieveText.SelectedText = regex.Replace(retrieveText.SelectedText, formBoxListBox.GetItemText(formBoxListBox.SelectedItem));
                }
            }
            else
            {
                ProcessErrors("Error!", 10, null, null);
            }

            retrieveText.Focus();
        }

        /*
        This resets [clears] the current selection in the condList.
        */
        private void condResetButton_Click(object sender, EventArgs e)
        {
            condList.ClearSelected();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            this.Text = Application.ProductName + " " + Application.ProductVersion;

            // Store the snippet file contents in memory.
            SetXmlDocument(xmlFile);

            // Set the default operating mode:
            string defaultMode = Properties.Settings.Default.DefaultMode;

            //MessageBox
            if (defaultMode.Contains("Ruby"))
            {
                MessageBox.Show("Switch to JS and make it default");
            }

            switch (defaultMode)
                        {


                            case ("retrieveJS"):

                                ModeSetRetJS();
                                prevRetrieveMode = retrieverMode.retrieveJS;
                                break;
        
            }
        
        }
        
        private void convertRubyToJSToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (retMode == retrieverMode.convertRubyToJS)
            {
                return;
            }

            ProcessMode(retrieverMode.convertRubyToJS);
        }
        
        private void retrieveRubyCodeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ProcessMode(retrieverMode.retrieveRuby);
        }
        
        private void retrieveJSCodeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ProcessMode(retrieverMode.retrieveJS);
        }

        private void setCurrentModeAsDefaultToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Properties.Settings.Default.DefaultMode = retMode.ToString();
            Properties.Settings.Default.Save();

            System.Configuration.ConfigurationManager.AppSettings["DefaultMode"] = retMode.ToString();
        }

        private void dataFlowCodeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Clipboard.SetText (this.retrieveText.Text.Trim());
        }

        


}
}
